# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from gravitino.api.rel.expressions.literals.literal import Literal
from gravitino.api.rel.types.types import Types
from gravitino.dto.rel.expressions.function_arg import FunctionArg

if TYPE_CHECKING:
    from gravitino.api.rel.types.type import Type


class LiteralDTO(Literal[str], FunctionArg):
    """Represents a Literal Data Transfer Object (DTO) that implements the Literal interface."""

    NULL: ClassVar[LiteralDTO]
    """An instance of LiteralDTO with a value of "NULL" and a data type of Types.NullType.get()."""

    def __init__(self, value: str, data_type: Type):
        self._value = value
        self._data_type = data_type

    def value(self) -> str:
        """The literal value.

        Returns:
            str: The value of the literal.
        """
        return self._value

    def data_type(self) -> Type:
        """The data type of the literal.

        Returns:
            Type: The data type of the literal.
        """
        return self._data_type

    def arg_type(self) -> FunctionArg.ArgType:
        return self.ArgType.LITERAL

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LiteralDTO):
            return False
        return (
            self._data_type == other.data_type()
            and self._value == other.value()
            and self.arg_type() == other.arg_type()
        )

    def __hash__(self) -> int:
        return hash((self.arg_type(), self._data_type, self._value))

    def __str__(self) -> str:
        return f"LiteralDTO(value='{self._value}', data_type={self._data_type})"

    @staticmethod
    def builder() -> Builder:
        """the builder for creating a new instance of LiteralDTO.

        Returns:
            Builder: the builder for creating a new instance of LiteralDTO.
        """
        return LiteralDTO.Builder()

    class Builder:
        """Builder for LiteralDTO."""

        def __init__(self):
            self._data_type = None
            self._value = None

        def with_value(self, value: str) -> LiteralDTO.Builder:
            """Set the value of the literal.

            Args:
                value (str): The value of the literal.

            Returns:
                Builder: The builder.
            """
            self._value = value
            return self

        def with_data_type(self, data_type: Type) -> LiteralDTO.Builder:
            """Set the data type of the literal.

            Args:
                data_type (Type): The data type of the literal.

            Returns:
                Builder: The builder.
            """
            self._data_type = data_type
            return self

        def build(self) -> LiteralDTO:
            """Builds a `LiteralDTO` instance.

            Returns:
                LiteralDTO: The `LiteralDTO` instance.
            """
            return LiteralDTO(value=self._value, data_type=self._data_type)


LiteralDTO.NULL = LiteralDTO("NULL", Types.NullType.get())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from gravitino.dto.responses.error_response import ErrorResponse


class ErrorHandler(ABC):
    """The ErrorHandler class is an abstract class specialized for handling ErrorResponse objects.
    Subclasses of ErrorHandler must implement the parseResponse method to provide custom parsing
    logic for different types of errors.
    """

    @abstractmethod
    def handle(self, error_response: ErrorResponse):
        """Handles the error response and raise the appropriate Exception. The implementation will use the
        provided error response to determine which exception to raise.

        Args:
          error_response: the error response from the server

        Raises:
          appropriate Exception.
        """



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from dataclasses_json import config

from gravitino.api.job.job_template_change import JobTemplateChange
from gravitino.dto.job.template_update_dto import TemplateUpdateDTO
from gravitino.rest.rest_message import RESTRequest


@dataclass
class JobTemplateUpdateRequest(RESTRequest, ABC):
    """Represents a request to update a job template."""

    _type: str = field(metadata=config(field_name="@type"))

    def __init__(self, action_type: str):
        self._type = action_type

    @abstractmethod
    def job_template_change(self) -> JobTemplateChange:
        """Converts the request to a JobTemplateChange object."""
        pass


@dataclass
class RenameJobTemplateRequest(JobTemplateUpdateRequest):
    """Request to rename a job template."""

    _new_name: str = field(metadata=config(field_name="newName"))
    """The new name for the job template."""

    def __init__(self, new_name: str):
        super().__init__("rename")
        self._new_name = new_name

    def validate(self):
        """Validates the fields of the request.

        Raises:
            ValueError if the new name is not set.
        """
        if not self._new_name:
            raise ValueError('"new_name" field is required and cannot be empty')

    def job_template_change(self) -> JobTemplateChange:
        return JobTemplateChange.rename(new_name=self._new_name)


@dataclass
class UpdateJobTemplateCommentRequest(JobTemplateUpdateRequest):
    """Request to update the comment of a job template."""

    _new_comment: str = field(metadata=config(field_name="newComment"))
    """The new comment for the job template."""

    def __init__(self, new_comment: str):
        super().__init__("updateComment")
        self._new_comment = new_comment

    def validate(self):
        """Validates the fields of the request.

        Raises:
            ValueError if the new comment is not set.
        """
        if self._new_comment is None:
            raise ValueError('"new_comment" field is required and cannot be None')

    def job_template_change(self) -> JobTemplateChange:
        return JobTemplateChange.update_comment(new_comment=self._new_comment)


@dataclass
class UpdateJobTemplateContentRequest(JobTemplateUpdateRequest):
    """Request to update the content of a job template."""

    _new_template: TemplateUpdateDTO = field(metadata=config(field_name="newTemplate"))
    """The template update details."""

    def __init__(self, new_template: TemplateUpdateDTO):
        super().__init__("updateTemplate")
        self._new_template = new_template

    def validate(self):
        """Validates the fields of the request.

        Raises:
            ValueError if the template update is not set.
        """
        if self._new_template is None:
            raise ValueError('"new_template" field is required and cannot be None')

    def job_template_change(self) -> JobTemplateChange:
        return JobTemplateChange.update_template(
            self._new_template.to_template_update()
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional


class MetadataObject(ABC):
    """The MetadataObject is the basic unit of the Gravitino system. It
    represents the metadata object in the Apache Gravitino system. The object
    can be a metalake, catalog, schema, table, topic, etc.
    """

    class Type(Enum):
        """The type of object in the Gravitino system. Every type will map one
        kind of the entity of the underlying system.
        """

        METALAKE = "metalake"
        """A metalake is a concept of tenant. It means an organization. A metalake
        contains many data sources.
        """

        CATALOG = "catalog"
        """A catalog is a collection of metadata from a specific metadata source,
        like Apache Hive catalog, Apache Iceberg catalog, JDBC catalog, etc.
        """

        SCHEMA = "schema"
        """"A schema is a sub collection of the catalog. The schema can contain filesets,
        tables, topics, etc.
        """

        FILESET = "fileset"
        """A fileset is mapped to a directory on a file system like HDFS, S3, ADLS,
        GCS, etc.
        """

        TABLE = "table"
        """A table is mapped the table of relational data sources like Apache Hive,
        MySQL, etc.
        """

        TOPIC = "topic"
        """A topic is mapped the topic of messaging data sources like Apache Kafka,
        Apache Pulsar, etc.
        """

        COLUMN = "column"
        """A column is a sub-collection of the table that represents a group of
        same type data.
        """

        ROLE = "role"
        """A role is an object contains specific securable objects with privileges."""

        MODEL = "model"
        """A model is mapped to the model artifact in ML."""

        TAG = "tag"
        """A tag is used to help manage other metadata object."""

        POLICY = "policy"
        """A policy can be associated with a metadata object for data governance and similar purposes."""

        JOB = "job"
        """A job represents a data processing task in Gravitino."""

        JOB_TEMPLATE = "job_template"
        """A job template represents a reusable template for creating jobs in Gravitino."""

    @abstractmethod
    def parent(self) -> Optional[str]:
        """The parent full name of the object.

        If the object doesn't have parent, this method will return `None`.

        Returns:
            Optional[str]: The parent full name of the object.
        """

    @abstractmethod
    def type(self) -> Type:
        """The type of the object.

        Returns:
            Type: The type of the object.
        """

    @abstractmethod
    def name(self) -> str:
        """The name of the object.

        Returns:
            str: The name of the object.
        """

    def full_name(self) -> str:
        """The full name of the object.

        Full name will be separated by "." to represent a string identifier of the object,
        like catalog, catalog.table, etc.

        Returns:
            str: The name of the object.
        """

        parent = self.parent()
        return ".".join([parent, self.name()]) if parent else self.name()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import config, dataclass_json

from gravitino.api.metadata_object import MetadataObject
from gravitino.utils.precondition import Precondition


@dataclass_json
@dataclass
class MetadataObjectDTO(MetadataObject):
    """Represents a Metadata Object DTO (Data Transfer Object)."""

    _type: MetadataObject.Type = field(metadata=config(field_name="type"))
    _full_name: str = field(metadata=config(field_name="fullName"))

    _name: str = field(init=False, default="")
    _parent: Optional[str] = field(init=False, default=None)

    def __post_init__(self) -> None:
        self.set_full_name(self._full_name)

    @staticmethod
    def builder() -> MetadataObjectDTO.Builder:
        return MetadataObjectDTO.Builder()

    def type(self) -> MetadataObject.Type:
        return self._type

    def parent(self) -> Optional[str]:
        return self._parent

    def name(self) -> str:
        return self._name

    def set_full_name(self, full_name: str) -> None:
        """
        Sets the full name of the metadata object.

        Args:
            full_name (str): The full name of the metadata object.
        """
        index = full_name.rfind(".")
        self._parent, self._name = (
            (None, full_name)
            if index == -1
            else (full_name[:index], full_name[index + 1 :])
        )

    class Builder:
        def __init__(self) -> None:
            self._full_name: str = ""
            self._type: MetadataObject.Type = None

        def full_name(self, full_name: str) -> MetadataObjectDTO.Builder:
            self._full_name = full_name
            return self

        def type(self, _type: MetadataObject.Type) -> MetadataObjectDTO.Builder:
            self._type = _type
            return self

        def build(self) -> MetadataObjectDTO:
            Precondition.check_string_not_empty(
                self._full_name, "Full name is required and cannot be empty."
            )
            Precondition.check_argument(
                self._type is not None, "Type is required and cannot be None."
            )

            return MetadataObjectDTO(
                self._type,
                self._full_name,
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import config

from gravitino.api.function.function_type import FunctionType
from gravitino.dto.function.function_definition_dto import FunctionDefinitionDTO
from gravitino.dto.function.function_dto import (
    _decode_function_type,
    _encode_function_type,
)
from gravitino.exceptions.base import IllegalArgumentException
from gravitino.rest.rest_message import RESTRequest


@dataclass
class FunctionRegisterRequest(RESTRequest):
    """Represents a request to register a function."""

    _name: str = field(metadata=config(field_name="name"))
    _function_type: FunctionType = field(
        metadata=config(
            field_name="functionType",
            encoder=_encode_function_type,
            decoder=_decode_function_type,
        )
    )
    _deterministic: bool = field(metadata=config(field_name="deterministic"))
    _definitions: List[FunctionDefinitionDTO] = field(
        metadata=config(field_name="definitions")
    )
    _comment: Optional[str] = field(default=None, metadata=config(field_name="comment"))

    def __init__(
        self,
        name: str,
        function_type: FunctionType,
        deterministic: bool,
        definitions: List[FunctionDefinitionDTO],
        comment: Optional[str] = None,
    ):
        self._name = name
        self._function_type = function_type
        self._deterministic = deterministic
        self._definitions = definitions
        self._comment = comment

    def validate(self):
        """Validates the request.

        Raises:
            IllegalArgumentException: If the request is invalid.
        """
        if not self._name:
            raise IllegalArgumentException(
                "'name' field is required and cannot be empty"
            )
        if self._function_type is None:
            raise IllegalArgumentException("'functionType' field is required")
        if not self._definitions:
            raise IllegalArgumentException(
                "'definitions' field is required and cannot be empty"
            )

        # Validate each definition has appropriate return type/columns based on function type
        for definition in self._definitions:
            if self._function_type == FunctionType.TABLE:
                return_columns = definition.return_columns()
                if not return_columns:
                    raise IllegalArgumentException(
                        "'returnColumns' is required in each definition for TABLE function type"
                    )
            else:
                if definition.return_type() is None:
                    raise IllegalArgumentException(
                        "'returnType' is required in each definition for SCALAR or AGGREGATE function type"
                    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from enum import Enum


class InternalClientType(Enum):
    """An enum class containing internal client type that supported."""

    HADOOP_GVFS = "HADOOP_GVFS"
    """The client type is `org.apache.gravitino.filesystem.hadoop.GravitinoVirtualFileSystem` which in
    the filesystem-hadoop3 module.
    """

    PYTHON_GVFS = "PYTHON_GVFS"
    """The client type is `gravitino.filesystem.gvfs.GravitinoVirtualFileSystem` which in the
    client-python module.
    """

    UNKNOWN = "UNKNOWN"
    """The client type is unknown.
    """



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import Enum


class FunctionType(Enum):
    """Function type supported by Gravitino."""

    SCALAR = "scalar"
    """Scalar function that returns a single value per row."""

    AGGREGATE = "aggregate"
    """Aggregate function that combines multiple rows into a single value."""

    TABLE = "table"
    """Table-valued function that returns a table of rows."""

    @classmethod
    def from_string(cls, type_str: str) -> "FunctionType":
        """Parse the function type from a string value.

        Args:
            type_str: The string to parse.

        Returns:
            The parsed FunctionType.

        Raises:
            ValueError: If the value cannot be parsed.
        """
        if not type_str:
            raise ValueError("Function type cannot be null or empty")

        type_lower = type_str.lower()
        if type_lower == "agg":
            return cls.AGGREGATE

        for func_type in cls:
            if func_type.value == type_lower:
                return func_type

        raise ValueError(f"Invalid function type: {type_str}")

    def type_name(self) -> str:
        """Returns the canonical string representation used by APIs."""
        return self.value



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import List

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.expressions.function_expression import FunctionExpression
from gravitino.dto.rel.expressions.function_arg import FunctionArg


class FuncExpressionDTO(FunctionExpression, FunctionArg):
    def __init__(self, function_name: str, function_args: List[FunctionArg]):
        self._function_name = function_name
        self._function_args = function_args

    def args(self) -> List[FunctionArg]:
        """The function arguments.

        Returns:
            List[FunctionArg]: The function arguments.
        """
        return self._function_args

    def function_name(self) -> str:
        return self._function_name

    def arguments(self) -> List[Expression]:
        return self._function_args

    def arg_type(self) -> FunctionArg.ArgType:
        return self.ArgType.FUNCTION

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FuncExpressionDTO):
            return False
        return (
            self._function_name == other.function_name()
            and self._function_args == other.args()
            and self.arg_type() is other.arg_type()
        )

    def __hash__(self) -> int:
        return hash((self.arg_type(), self._function_name, tuple(self._function_args)))

    @staticmethod
    def builder() -> Builder:
        """The builder for creating a new instance of `FuncExpressionDTO`.

        Returns:
            Builder: The builder for creating a new instance of `FuncExpressionDTO`.
        """
        return FuncExpressionDTO.Builder()

    class Builder:
        """Builder for `FuncExpressionDTO.`"""

        def __init__(self):
            self._function_args = None
            self._function_name = None

        def with_function_name(self, function_name: str) -> FuncExpressionDTO.Builder:
            """Set the function name for the function expression.

            Args:
                function_name (str): The function name.

            Returns:
                FuncExpressionDTO.Builder: The builder.
            """

            self._function_name = function_name
            return self

        def with_function_args(
            self, function_args: List[FunctionArg]
        ) -> FuncExpressionDTO.Builder:
            """Set the function arguments for the function expression.

            Args:
                function_args (List[FunctionArg]): The function arguments.

            Returns:
                FuncExpressionDTO.Builder: The builder.
            """

            self._function_args = function_args
            return self

        def build(self) -> FuncExpressionDTO:
            """Build the function expression.

            Returns:
                FuncExpressionDTO: The function expression.
            """

            return FuncExpressionDTO(
                function_name=self._function_name, function_args=self._function_args
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC


class ModelVersionChange(ABC):
    """
    A model version change is a change to a model version. It can be used to update the comment
    of a model version, set a property and value pair for a model version, or remove a property from a model.
    """

    @staticmethod
    def update_comment(comment: str):
        """Creates a new model version change to update the comment of the model version.
        Args:
            comment: The new comment of the model version.
        Returns:
            The model version change.
        """
        return ModelVersionChange.UpdateComment(comment)

    @staticmethod
    def set_property(key, value):
        """Creates a new model version change to set a property and value pair for the model version.
        Args:
            key: The key of the property.
            value: The value of the property.
        Returns:
            The model version change.
        """
        return ModelVersionChange.SetProperty(key, value)

    @staticmethod
    def remove_property(key: str):
        """Creates a new model version change to remove a property from the model version.
        Args:
            key: The key of the property.
        Returns:
            The model version change.
        """
        return ModelVersionChange.RemoveProperty(key)

    @staticmethod
    def update_uri(uri: str, uri_name: str = None):
        """Creates a new model version change to update the uri of the model version.
        Args:
            uri: The new uri of the model version.
            uri_name: The uri name of the model version to be updated.
        Returns:
            The model version change.
        """
        return ModelVersionChange.UpdateUri(uri, uri_name)

    @staticmethod
    def add_uri(uri_name: str, uri: str):
        """Creates a new model version change to add the uri of the model version.
        Args:
            uri_name: The uri name of the model version to be added.
            uri: The new uri of the model version to be added.
        Returns:
            The model version change.
        """
        return ModelVersionChange.AddUri(uri_name, uri)

    @staticmethod
    def remove_uri(uri_name: str):
        """Creates a new model version change to remove the uri of the model version.
        Args:
            uri_name: The uri name of the model version to be removed.
        Returns:
            The model version change.
        """
        return ModelVersionChange.RemoveUri(uri_name)

    @staticmethod
    def update_aliases(aliases_to_add, aliases_to_remove):
        """Creates a new model version change to update the aliases of the model version.
        Args:
            aliases_to_add: The new aliases to add to the model version.
            aliases_to_remove: The aliases to remove from the model version.
        Returns:
            The model version change.
        """
        return ModelVersionChange.UpdateAliases(aliases_to_add, aliases_to_remove)

    class UpdateComment:
        """A model version change to update the comment of the model version."""

        def __init__(self, new_comment: str):
            self._new_comment = new_comment

        def new_comment(self) -> str:
            """Retrieves the new comment of the model version.
            Returns:
                The new comment of the model version.
            """
            return self._new_comment

        def __eq__(self, other):
            """Compares this UpdateComment instance with another object for equality. Two instances are
            considered equal if they designate the same new comment for the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version comment update operation;
                false otherwise.
            """
            if not isinstance(other, ModelVersionChange.UpdateComment):
                return False

            return self.new_comment() == other.new_comment()

        def __hash__(self):
            """Generates a hash code for this UpdateComment instance. The hash code is primarily based on
            the new comment for the model version.
            Returns:
                A hash code value for this comment update operation.
            """
            return hash(self.new_comment())

        def __str__(self):
            """Provides a string representation of the UpdateComment instance. This string includes the
            class name followed by the new comment of the model version.
            Returns:
                A string summary of this comment update operation.
            """
            return f"UpdateComment {self._new_comment}"

    class SetProperty:
        """A model version change to set a property and value pair for the model version."""

        def __init__(self, key: str, value: str):
            self._key = key
            self._value = value

        def property(self) -> str:
            """Retrieves the name of the property.
            Returns:
                The name of the property.
            """
            return self._key

        def value(self) -> str:
            """Retrieves the value of the property.
            Returns:
                The value of the property.
            """
            return self._value

        def __eq__(self, other):
            """Compares this SetProperty instance with another object for equality. Two instances are
            considered equal if they designate the same key and value pair for the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version property set operation;  false otherwise.
            """
            if not isinstance(other, ModelVersionChange.SetProperty):
                return False
            return self.key() == other.key() and self.value() == other.value()

        def __hash__(self):
            """Generates a hash code for this SetProperty instance. The hash code is primarily based on
            the key and value pair for the model version.
            Returns:
                A hash code value for this property set operation.
            """
            return hash((self.key(), self.value()))

        def __str__(self):
            """Provides a string representation of the SetProperty instance. This string includes the
            class name followed by the key and value pair for the model version.
            Returns:
                A string summary of this property set operation.
            """
            return f"SetProperty {self.key()}={self.value()}"

    class RemoveProperty:
        """A model version change to remove a property from the model version."""

        def __init__(self, key: str):
            self._key = key

        def property(self) -> str:
            """Retrieves the name of the property.
            Returns:
                The name of the property.
            """
            return self._key

        def __eq__(self, other):
            """Compares this RemoveProperty instance with another object for equality. Two instances are
            considered equal if they designate the same key for the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version property remove operation;
                false otherwise.
            """
            if not isinstance(other, ModelVersionChange.RemoveProperty):
                return False
            return self.key() == other.key()

        def __hash__(self):
            """Generates a hash code for this RemoveProperty instance. The hash code is primarily based on
            the key for the model version.
            Returns:
                A hash code value for this property remove operation.
            """
            return hash(self.key())

        def __str__(self):
            """Provides a string representation of the RemoveProperty instance. This string includes the
            class name followed by the key for the model version.
            Returns:
                A string summary of this property remove operation.
            """
            return f"RemoveProperty {self.key()}"

    class UpdateUri:
        """A model version change to update the URI of the model version."""

        def __init__(self, new_uri: str, uri_name: str = None):
            self._new_uri = new_uri
            self._uri_name = uri_name

        def new_uri(self) -> str:
            """Retrieves the new URI of the model version.
            Returns:
                The new URI of the model version.
            """
            return self._new_uri

        def uri_name(self) -> str:
            """Retrieves the URI name of the model version to be updated.
            Returns:
                The URI name of the model version to be updated.
            """
            return self._uri_name

        def __eq__(self, other):
            """Compares this UpdateUri instance with another object for equality. Two instances are
            considered equal if they designate the same new URI and URI name for the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version URI update operation;
                false otherwise.
            """
            if not isinstance(other, ModelVersionChange.UpdateUri):
                return False
            return (
                self.new_uri() == other.new_uri()
                and self.uri_name() == other.uri_name()
            )

        def __hash__(self):
            """Generates a hash code for this UpdateUri instance. The hash code is primarily based on
            the new URI and its name for the model version.
            Returns:
                A hash code value for this URI update operation.
            """
            return hash((self.new_uri(), self.uri_name()))

        def __str__(self):
            """Provides a string representation of the UpdateUri instance. This string includes the
            class name followed by the new URI and its name of the model version.
            Returns:
                A string summary of this URI update operation.
            """
            return f"UpdateUri uriName: ({self._uri_name}) newUri: ({self._new_uri})"

    class AddUri:
        """A ModelVersionChange to add a uri of the model version."""

        def __init__(self, uri_name: str, uri: str):
            self._uri_name = uri_name
            self._uri = uri

        def uri_name(self) -> str:
            """Retrieves the URI name of the model version to be added.
            Returns:
                The URI name of the model version to be added.
            """
            return self._uri_name

        def uri(self) -> str:
            """Retrieves the URI of the model version to be added.
            Returns:
                The new URI of the model version to be added.
            """
            return self._uri

        def __eq__(self, other):
            """Compares this AddUri instance with another object for equality. Two instances are
            considered equal if they designate the same URI and URI name for the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version URI add operation;
                false otherwise.
            """
            if not isinstance(other, ModelVersionChange.AddUri):
                return False
            return self.uri_name() == other.uri_name() and self.uri() == other.uri()

        def __hash__(self):
            """Generates a hash code for this AddUri instance. The hash code is primarily based on
            the URI and its name for the model version.
            Returns:
                A hash code value for this URI add operation.
            """
            return hash((self.uri_name(), self.uri()))

        def __str__(self):
            """Provides a string representation of the AddUri instance. This string includes the
            class name followed by the URI and its name of the model version.
            Returns:
                A string summary of this URI add operation.
            """
            return f"AddUri uriName: ({self._uri_name}) uri: ({self.uri})"

    class RemoveUri:
        """A ModelVersionChange to remove a uri of the model version."""

        def __init__(self, uri_name: str):
            self._uri_name = uri_name

        def uri_name(self) -> str:
            """Retrieves the URI name of the model version to be removed.
            Returns:
                The URI name of the model version to be removed.
            """
            return self._uri_name

        def __eq__(self, other):
            """Compares this RemoveUri instance with another object for equality. Two instances are
            considered equal if they designate the same URI name for the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version URI remove operation;
                false otherwise.
            """
            if not isinstance(other, ModelVersionChange.RemoveUri):
                return False
            return self.uri_name() == other.uri_name()

        def __hash__(self):
            """Generates a hash code for this RemoveUri instance. The hash code is primarily based on
            the URI name for the model version.
            Returns:
                A hash code value for this URI remove operation.
            """
            return hash(self.uri_name())

        def __str__(self):
            """Provides a string representation of the RemoveUri instance. This string includes the
            class name followed by the URI name of the model version.
            Returns:
                A string summary of this URI remove operation.
            """
            return f"RemoveUri uriName: ({self._uri_name})"

    class UpdateAliases:
        """A model version change to update the aliases of the model version."""

        def __init__(self, aliases_to_add, aliases_to_remove):
            self._aliases_to_add = set(aliases_to_add) if aliases_to_add else set()
            self._aliases_to_remove = (
                set(aliases_to_remove) if aliases_to_remove else set()
            )

        def aliases_to_add(self):
            """Retrieves the aliases to add.
            Returns:
                The aliases to add.
            """
            return self._aliases_to_add

        def aliases_to_remove(self):
            """Retrieves the aliases to remove.
            Returns:
                The aliases to remove.
            """
            return self._aliases_to_remove

        def __eq__(self, other) -> bool:
            """Compares this UpdateAliases instance with another object for equality. Two instances are
            considered equal if they designate the same aliases to add and remove from the model version.
            Args:
                other: The object to compare with this instance.
            Returns:
                true if the given object represents an identical model version alias update operation;
                false otherwise.
            """
            if not isinstance(other, ModelVersionChange.UpdateAliases):
                return False
            return (
                self.aliases_to_add() == other.aliases_to_add()
                and self.aliases_to_remove() == other.aliases_to_remove()
            )

        def __hash__(self) -> int:
            """Generates a hash code for this UpdateAliases instance. The hash code is primarily based on
            the aliases to add and remove from the model version.
            Returns:
                A hash code value for this alias update operation.
            """
            return hash((self._aliases_to_add, self._aliases_to_remove))

        def __str__(self) -> str:
            """Returns a string representation of the UpdateAliases instance. This string includes the
            class name followed by the aliases to add and remove for the model version.
            Returns:
                A string summary of this alias update operation.
            """
            add_str = ", ".join(sorted(self._aliases_to_add))
            remove_str = ", ".join(sorted(self._aliases_to_remove))
            return f"UpdateAlias AliasToAdd: ({add_str}) AliasToRemove: ({remove_str})"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from abc import abstractmethod, ABC
from typing import List, Dict, Optional, TypeVar, Generic, Any
from enum import Enum
from copy import deepcopy


class JobType(Enum):
    """
    JobType is an enum to define the type of the job.

    Gravitino supports different types of jobs, such as Spark and Shell. The job type is
    required to determine the runtime environment for executing the job.
    """

    SPARK = "spark"
    """job type for executing Spark jobs"""
    SHELL = "shell"
    """job type for executing shell commands"""

    @classmethod
    def job_type_serialize(cls, job_type: "JobType") -> str:
        """
        Serializes the JobType to a string.

        Args:
            job_type: The JobType to serialize.

        Returns:
            The serialized string representation of the JobType.
        """
        return job_type.value.lower()

    @classmethod
    def job_type_deserialize(cls, job_type_str: str) -> "JobType":
        """
        Deserializes a string to a JobType.

        Args:
            job_type_str: The string representation of the JobType.

        Returns:
            The JobType corresponding to the string.
        """
        for m in cls:
            if m.value.lower() == job_type_str.lower():
                return m

        # If no match found, raise an error
        raise ValueError(f"Invalid job type string: {job_type_str}")


T = TypeVar("T", bound="JobTemplate")
B = TypeVar("B", bound="BaseBuilder")


class JobTemplate(ABC):
    """
    JobTemplate is a class to define all the configuration parameters for a job.

    JobType is enum to define the type of the job, because Gravitino needs different runtime
    environments to execute different types of jobs, so the job type is required.

    Some parameters can be templated, which means that they can be replaced with actual values
    when running the job, for example, arguments can be { "{input_path}}", "{{output_path}" },
    environment variables can be { "foo": "{{foo_value}}", "bar": "{{bar_value}}" }. the parameters
    support templating are: arguments, environments.
    """

    def __init__(self, builder: "BaseBuilder"):
        self._name = builder.name
        self._comment = builder.comment
        self._executable = builder.executable
        self._arguments = builder.arguments
        self._environments = builder.environments
        self._custom_fields = builder.custom_fields

    @property
    def name(self) -> str:
        """
        Returns:
            The name of the job template.
        """
        return self._name

    @property
    def comment(self) -> Optional[str]:
        """
        Returns:
            The comment of the job template, which can be used to describe the job.
            This field is optional and can be None.
        """
        return self._comment

    @property
    def executable(self) -> str:
        """
        Returns:
            The executable of the job template, which is the command or script that will be executed
            when the job is run. This field is required and must not be empty.
        """
        return self._executable

    @property
    def arguments(self) -> List[str]:
        """
        Returns:
            The arguments of the job template, which are the parameters that will be passed to the
            executable when the job is run.
        """
        return self._arguments

    @property
    def environments(self) -> Dict[str, str]:
        """
        Returns:
            The environment variables of the job template, which are key-value pairs that will be
            set in the environment when the job is run. This can include variables that can be
            templated with actual values when running the job.
        """
        return self._environments

    @property
    def custom_fields(self) -> Dict[str, str]:
        """
        Returns:
            Custom fields of the job template, which are additional key-value pairs that can be
            used to store any other information related to the job. This field is optional
            and can be empty.
        """
        return self._custom_fields

    @abstractmethod
    def job_type(self) -> JobType:
        """
        Returns:
            The type of the job template, which is an enum value indicating the type of job
            (e.g., SPARK, SHELL). This is required to determine the runtime environment for executing
            the job.
        """
        pass

    def __eq__(self, other: Any) -> bool:
        if self is other:
            return True
        if not isinstance(other, JobTemplate):
            return False
        return (
            self.job_type() == other.job_type()
            and self._name == other.name
            and self._comment == other.comment
            and self._executable == other.executable
            and self._arguments == other.arguments
            and self._environments == other.environments
            and self._custom_fields == other.custom_fields
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.job_type(),
                self._name,
                self._comment,
                self._executable,
                tuple(self._arguments),
                frozenset(self._environments.items()),
                frozenset(self._custom_fields.items()),
            )
        )

    def __str__(self) -> str:
        res = f"  jobType={self.job_type()},\n  name='{self._name}',\n"
        if self._comment:
            res += f"  comment='{self._comment}',\n"
        res += f"  executable='{self._executable}',\n"

        if self._arguments:
            res += (
                "  arguments=[\n"
                + "".join(f"    {arg},\n" for arg in self._arguments)
                + "  ],\n"
            )
        else:
            res += "  arguments=[],\n"

        if self._environments:
            res += (
                "  environments={\n"
                + "".join(f"    {k}: {v},\n" for k, v in self._environments.items())
                + "  },\n"
            )
        else:
            res += "  environments={},\n"

        if self._custom_fields:
            res += (
                "  customFields={\n"
                + "".join(f"    {k}: {v},\n" for k, v in self._custom_fields.items())
                + "  }\n"
            )
        else:
            res += "  customFields={}\n"

        return res

    class BaseBuilder(Generic[B, T]):
        """
        BaseBuilder is a generic class to build a JobTemplate.
        """

        def __init__(self: B):
            self.name: Optional[str] = None
            self.comment: Optional[str] = None
            self.executable: Optional[str] = None
            self.arguments: List[str] = []
            self.environments: Dict[str, str] = {}
            self.custom_fields: Dict[str, str] = {}

        def with_name(self: B, name: str) -> B:
            """
            Sets the name of the job template.

            Args:
                name: The name of the job template. It must not be null or empty.

            Returns:
                The builder instance for method chaining.
            """
            self.name = name
            return self

        def with_comment(self: B, comment: str) -> B:
            """
            Sets the comment of the job template.

            Args:
                comment: The comment of the job template. It can be null or empty.

            Returns:
                The builder instance for method chaining.
            """
            self.comment = comment
            return self

        def with_executable(self: B, executable: str) -> B:
            """
            Sets the executable of the job template.

            Args:
                executable: The executable of the job template. It must not be null or empty.

            Returns:
                The builder instance for method chaining.
            """
            self.executable = executable
            return self

        def with_arguments(self: B, arguments: List[str]) -> B:
            """
            Sets the arguments of the job template.

            Args:
                arguments: The arguments of the job template.

            Returns:
                The builder instance for method chaining.
            """
            self.arguments = deepcopy(arguments)
            return self

        def with_environments(self: B, environments: Dict[str, str]) -> B:
            """
            Sets the environment variables of the job template.

            Args:
                environments: The environment variables of the job template.

            Returns:
                The builder instance for method chaining.
            """
            self.environments = deepcopy(environments)
            return self

        def with_custom_fields(self: B, custom_fields: Dict[str, str]) -> B:
            """
            Sets the custom fields of the job template.

            Args:
                custom_fields: The custom fields of the job template.

            Returns:
                The builder instance for method chaining.
            """
            self.custom_fields = deepcopy(custom_fields)
            return self

        def validate(self: B):
            """
            Validates the job template parameters.
            """
            if not self.name or not self.name.strip():
                raise ValueError("Job name must not be null or empty")
            if not self.executable or not self.executable.strip():
                raise ValueError("Executable must not be null or empty")
            self.arguments = self.arguments or []
            self.environments = self.environments or {}
            self.custom_fields = self.custom_fields or {}

        @abstractmethod
        def build(self) -> T:
            """
            Builds the JobTemplate instance.

            Returns:
                An instance of JobTemplate with the configured parameters.
            """
            pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import List, Optional

from gravitino.api.function.function import Function
from gravitino.api.function.function_change import FunctionChange
from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_type import FunctionType
from gravitino.exceptions.base import NoSuchFunctionException
from gravitino.name_identifier import NameIdentifier
from gravitino.namespace import Namespace


class FunctionCatalog(ABC):
    """The FunctionCatalog interface defines the public API for managing functions in a schema."""

    @abstractmethod
    def list_functions(self, namespace: Namespace) -> List[NameIdentifier]:
        """List the functions in a namespace from the catalog.

        Args:
            namespace: A namespace.

        Returns:
            A list of function identifiers in the namespace.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
        """
        pass

    @abstractmethod
    def list_function_infos(self, namespace: Namespace) -> List[Function]:
        """List the functions with details in a namespace from the catalog.

        Args:
            namespace: A namespace.

        Returns:
            A list of functions in the namespace.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
        """
        pass

    @abstractmethod
    def get_function(self, ident: NameIdentifier) -> Function:
        """Get a function by NameIdentifier from the catalog.

        The identifier only contains the schema and function name. A function may
        include multiple definitions (overloads) in the result.

        Args:
            ident: A function identifier.

        Returns:
            The function with the given name.

        Raises:
            NoSuchFunctionException: If the function does not exist.
        """
        pass

    def function_exists(self, ident: NameIdentifier) -> bool:
        """Check if a function with the given name exists in the catalog.

        Args:
            ident: The function identifier.

        Returns:
            True if the function exists, False otherwise.
        """
        try:
            self.get_function(ident)
            return True
        except NoSuchFunctionException:
            return False

    @abstractmethod
    def register_function(
        self,
        ident: NameIdentifier,
        comment: Optional[str],
        function_type: FunctionType,
        deterministic: bool,
        definitions: List[FunctionDefinition],
    ) -> Function:
        """Register a function with one or more definitions (overloads).

        Each definition contains its own return type (for scalar/aggregate functions)
        or return columns (for table-valued functions).

        Args:
            ident: The function identifier.
            comment: The optional function comment.
            function_type: The function type (SCALAR, AGGREGATE, or TABLE).
            deterministic: Whether the function is deterministic.
            definitions: The function definitions, each containing parameters,
                return type/columns, and implementations.

        Returns:
            The registered function.

        Raises:
            NoSuchSchemaException: If the schema does not exist.
            FunctionAlreadyExistsException: If the function already exists.
        """
        pass

    @abstractmethod
    def alter_function(
        self, ident: NameIdentifier, *changes: FunctionChange
    ) -> Function:
        """Applies FunctionChange changes to a function in the catalog.

        Implementations may reject the changes. If any change is rejected, no changes
        should be applied to the function.

        Args:
            ident: The NameIdentifier instance of the function to alter.
            changes: The FunctionChange instances to apply to the function.

        Returns:
            The updated Function instance.

        Raises:
            NoSuchFunctionException: If the function does not exist.
            IllegalArgumentException: If the change is rejected by the implementation.
        """
        pass

    @abstractmethod
    def drop_function(self, ident: NameIdentifier) -> bool:
        """Drop a function by name.

        Args:
            ident: The name identifier of the function.

        Returns:
            True if the function is deleted, False if the function does not exist.
        """
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import List, Optional


class FunctionResources:
    """Represents external resources that are required by a function implementation."""

    def __init__(
        self,
        jars: Optional[List[str]] = None,
        files: Optional[List[str]] = None,
        archives: Optional[List[str]] = None,
    ):
        """Create a FunctionResources instance.

        Args:
            jars: The jar resources.
            files: The file resources.
            archives: The archive resources.
        """
        self._jars = list(jars) if jars else []
        self._files = list(files) if files else []
        self._archives = list(archives) if archives else []

    @classmethod
    def empty(cls) -> "FunctionResources":
        """Returns an empty FunctionResources instance."""
        return cls()

    @classmethod
    def of(
        cls,
        jars: Optional[List[str]] = None,
        files: Optional[List[str]] = None,
        archives: Optional[List[str]] = None,
    ) -> "FunctionResources":
        """Create a FunctionResources instance.

        Args:
            jars: The jar resources.
            files: The file resources.
            archives: The archive resources.

        Returns:
            A FunctionResources instance.
        """
        return cls(jars, files, archives)

    def jars(self) -> List[str]:
        """Returns the jar resources."""
        return list(self._jars)

    def files(self) -> List[str]:
        """Returns the file resources."""
        return list(self._files)

    def archives(self) -> List[str]:
        """Returns the archive resources."""
        return list(self._archives)

    def __eq__(self, other) -> bool:
        if not isinstance(other, FunctionResources):
            return False
        return (
            self._jars == other._jars
            and self._files == other._files
            and self._archives == other._archives
        )

    def __hash__(self) -> int:
        return hash((tuple(self._jars), tuple(self._files), tuple(self._archives)))

    def __repr__(self) -> str:
        return (
            f"FunctionResources(jars={self._jars}, "
            f"files={self._files}, archives={self._archives})"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod

from dataclasses_json import DataClassJsonMixin


class RESTMessage(DataClassJsonMixin, ABC):
    """
    Interface for REST messages.

    REST messages are objects that are sent to and received from REST endpoints. They are
    typically used to represent the request and response bodies of REST API calls.
    """

    @abstractmethod
    def validate(self):
        """
        Ensures that a constructed instance of a REST message is valid according to the REST spec.

        This is needed when parsing data that comes from external sources and the object might have
        been constructed without all the required fields present.

        Raises:
            IllegalArgumentException: If the message is not valid.
        """
        pass


class RESTRequest(RESTMessage, ABC):
    """Interface to mark a REST request."""


class RESTResponse(RESTMessage, ABC):
    """Interface to mark a REST response"""



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional, cast, final

from dataclasses_json import config

from gravitino.api.rel.column import Column
from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.types.type import Type


class TableChange(ABC):
    """Defines the public APIs for managing tables in a schema.

    The `TableChange` interface defines the public API for managing tables in a schema.
    If the catalog implementation supports tables, it must implement this interface.
    """

    @staticmethod
    def rename(new_name: str) -> "RenameTable":
        """Create a `TableChange` for renaming a table.

        Args:
            new_name: The new table name.

        Returns:
            RenameTable: A `TableChange` for the rename.
        """
        return TableChange.RenameTable(new_name)

    @staticmethod
    def update_comment(new_comment: str) -> "UpdateComment":
        """Create a `TableChange` for updating the comment.

        Args:
            new_comment: The new comment.

        Returns:
            UpdateComment: A `TableChange` for the update.
        """
        return TableChange.UpdateComment(new_comment)

    @staticmethod
    def set_property(property_name: str, value: str) -> "SetProperty":
        """Create a `TableChange` for setting a table property.

        If the property already exists, it will be replaced with the new value.

        Args:
            property_name (str): The property name.
            value (str): The new property value.

        Returns:
            SetProperty: A `TableChange` for the addition.
        """
        return TableChange.SetProperty(property_name, value)

    @staticmethod
    def remove_property(property_name: str) -> "RemoveProperty":
        """Create a `TableChange` for removing a table property.

        If the property does not exist, the change will succeed.

        Args:
            property_name (str): The property name.

        Returns:
            RemoveProperty: A `TableChange` for the addition.
        """
        return TableChange.RemoveProperty(property_name)

    @staticmethod
    def add_column(
        field_name: list[str],
        data_type: Type,
        comment: Optional[str] = None,
        position: Optional["TableChange.ColumnPosition"] = None,
        nullable: bool = True,
        auto_increment: bool = False,
        default_value: Optional[Expression] = None,
    ) -> "AddColumn":
        """Create a `TableChange` for adding a column.

        Args:
            field_name (list[str]):
                Field name of the new column.
            data_type (Type):
                The new column's data type.
            comment (Optional[str]):
                The new field's comment string, defaults to `None`.
            position (Optional[TableChange.ColumnPosition]):
                The new column's position, defaults to `None`.
            nullable (bool):
                The new column's nullable.
            auto_increment (bool):
                The new column's autoIncrement.
            default_value (Expression):
                The new column's default value.

        Returns:
            AddColumn: A `TableChange` for the addition.
        """
        return AddColumn(
            field_name,
            data_type,
            comment,
            position,
            nullable,
            auto_increment,
            default_value,
        )

    @staticmethod
    def rename_column(field_name: list[str], new_name: str) -> "RenameColumn":
        """Create a `TableChange` for renaming a field.

        The name is used to find the field to rename. The new name will replace the **leaf field
        name**. For example, `rename_column(["a", "b", "c"], "x")` should produce column **a.b.x**.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The current field name.
            new_name (str): The new name.

        Returns:
            RenameColumn: A TableChange for the rename.
        """
        return RenameColumn(field_name, new_name)

    @staticmethod
    def update_column_default_value(
        field_name: list[str], new_default_value: Expression
    ) -> "UpdateColumnDefaultValue":
        """Create a `TableChange` for updating the default value of a field.

        The name is used to find the field to update.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The field name of the column to update.
            new_default_value (Expression): The new default value.

        Returns:
            TableChange: A `TableChange` for the update.
        """
        return UpdateColumnDefaultValue(field_name, new_default_value)

    @staticmethod
    def update_column_type(
        field_name: list[str], new_data_type: Type
    ) -> "UpdateColumnType":
        """Create a `TableChange` for updating the type of a field that is nullable.

        The field name are used to find the field to update.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The field name of the column to update.
            new_data_type (Type): The new data type.

        Returns:
            TableChange: A `TableChange` for the update.
        """
        return UpdateColumnType(field_name, new_data_type)

    @staticmethod
    def update_column_comment(
        field_name: list[str], new_comment: str
    ) -> "UpdateColumnComment":
        """Create a `TableChange` for updating the comment of a field.

        The name is used to find the field to update.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The field name of the column to update.
            new_comment (str): The new comment.

        Returns:
            TableChange: A `TableChange` for the update.
        """
        return UpdateColumnComment(field_name, new_comment)

    @staticmethod
    def update_column_position(
        field_name: list[str], new_position: "TableChange.ColumnPosition"
    ) -> "UpdateColumnPosition":
        """Create a `TableChange` for updating the position of a field.

        The name is used to find the field to update.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The field name of the column to update.
            new_position (TableChange.ColumnPosition): The new position.

        Returns:
            TableChange: A `TableChange` for the update.
        """
        return UpdateColumnPosition(field_name, new_position)

    @staticmethod
    def delete_column(field_name: list[str], if_exists: bool) -> "DeleteColumn":
        """Create a `TableChange` for deleting a field.

        If the field does not exist, the change will result in an `IllegalArgumentException`
        unless `if_exists` is true.

        Args:
            field_name (list[str]): Field name of the column to delete.
            if_exists (bool): If true, silence the error if column does not exist during drop.
                Otherwise, an `IllegalArgumentException` will be thrown.

        Returns:
            TableChange: A `TableChange` for the delete.
        """
        return DeleteColumn(field_name, if_exists)

    @staticmethod
    def update_column_nullability(
        field_name: list[str], nullable: bool
    ) -> "UpdateColumnNullability":
        """Create a `TableChange` for updating the nullability of a field.

        The name is used to find the field to update.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The field name of the column to update.
            nullable (bool): The new nullability.

        Returns:
            TableChange: A `TableChange` for the update.
        """
        return UpdateColumnNullability(field_name, nullable)

    @staticmethod
    def add_index(
        index_type: Index.IndexType,
        name: str,
        field_names: list[list[str]],
    ) -> "AddIndex":
        """Create a `TableChange` for adding an index.

        Args:
            index_type (Index.IndexType): The type of the index.
            name (str): The name of the index.
            field_names (list[list[str]]): The field names of the index.

        Returns:
            TableChange: A `TableChange` for the add index.
        """
        return TableChange.AddIndex(index_type, name, field_names)

    @staticmethod
    def delete_index(name: str, if_exists: bool) -> "DeleteIndex":
        """Create a `TableChange` for deleting an index.

        Args:
            name (str): The name of the index to be dropped.
            if_exists (bool): If true, silence the error if column does not exist during drop.
                Otherwise, an `IllegalArgumentException` will be thrown.

        Returns:
            TableChange: A `TableChange` for the delete index.
        """
        return TableChange.DeleteIndex(name, if_exists)

    @staticmethod
    def update_column_auto_increment(
        field_name: list[str], auto_increment: bool
    ) -> "UpdateColumnAutoIncrement":
        """Create a `TableChange` for updating the autoIncrement of a field.

        The name is used to find the field to update.

        If the field does not exist, the change will result in an `IllegalArgumentException`.

        Args:
            field_name (list[str]): The field name of the column to update.
            auto_increment (bool): The new autoIncrement.

        Returns:
            TableChange: A `TableChange` for the update.
        """
        return UpdateColumnAutoIncrement(field_name, auto_increment)

    @final
    @dataclass(frozen=True)
    class RenameTable:
        """A `TableChange` to rename a table."""

        _new_name: str = field(metadata=config(field_name="new_name"))

        def get_new_name(self) -> str:
            """Retrieves the new name for the table.

            Returns:
                str: The new name of the table.
            """
            return self._new_name

        def __str__(self):
            return f"RENAMETABLE {self._new_name}"

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, TableChange.RenameTable):
                return False
            other = cast(TableChange.RenameTable, value)
            return self._new_name == other.get_new_name()

        def __hash__(self) -> int:
            return hash(self._new_name)

    @final
    @dataclass(frozen=True)
    class UpdateComment:
        """A `TableChange` to update a table's comment."""

        _new_comment: str = field(metadata=config(field_name="new_comment"))

        def get_new_comment(self) -> str:
            """Retrieves the new comment for the table.

            Returns:
                str: The new comment of the table.
            """
            return self._new_comment

        def __str__(self):
            return f"UPDATECOMMENT {self._new_comment}"

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, TableChange.UpdateComment):
                return False
            other = cast(TableChange.UpdateComment, value)
            return self._new_comment == other.get_new_comment()

        def __hash__(self) -> int:
            return hash(self._new_comment)

    @final
    @dataclass(frozen=True)
    class SetProperty:
        """A `TableChange` to set a table property."""

        _property: str = field(metadata=config(field_name="property"))
        _value: str = field(metadata=config(field_name="value"))

        def get_property(self) -> str:
            """Retrieves the name of the property.

            Returns:
                str: The name of the property.
            """
            return self._property

        def get_value(self) -> str:
            """Retrieves the value of the property.

            Returns:
                str: The value of the property.
            """
            return self._value

        def __str__(self):
            return f"SETPROPERTY {self._property} {self._value}"

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, TableChange.SetProperty):
                return False
            other = cast(TableChange.SetProperty, value)
            return (
                self._property == other.get_property()
                and self._value == other.get_value()
            )

        def __hash__(self) -> int:
            return hash((self._property, self._value))

    @final
    @dataclass(frozen=True)
    class RemoveProperty:
        """A `TableChange` to remove a table property.

        If the property does not exist, the change should succeed.
        """

        _property: str = field(metadata=config(field_name="property"))

        def get_property(self) -> str:
            """Retrieves the name of the property to be removed from the table.

            Returns:
                str: The name of the property scheduled for removal.
            """
            return self._property

        def __str__(self):
            return f"REMOVEPROPERTY {self._property}"

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, TableChange.RemoveProperty):
                return False
            other = cast(TableChange.RemoveProperty, value)
            return self._property == other.get_property()

        def __hash__(self) -> int:
            return hash(self._property)

    @final
    @dataclass(frozen=True)
    class AddIndex:
        """A TableChange to add an index.

        Add an index key based on the type and field name passed in as well as the name.
        """

        _type: Index.IndexType = field(metadata=config(field_name="type"))
        _name: str = field(metadata=config(field_name="name"))
        _field_names: list[list[str]] = field(metadata=config(field_name="field_names"))

        def get_type(self) -> Index.IndexType:
            """Retrieves the type of the index.

            Returns:
                IndexType: The type of the index.
            """
            return self._type

        def get_name(self) -> str:
            """Retrieves the name of the index.

            Returns:
                str: The name of the index.
            """
            return self._name

        def get_field_names(self) -> list[list[str]]:
            """Retrieves the field names of the index.

            Returns:
                list[list[str]]: The field names of the index.
            """
            return self._field_names

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, TableChange.AddIndex):
                return False
            other = cast(TableChange.AddIndex, value)
            return (
                self._type == other.get_type()
                and self._name == other.get_name()
                and self._field_names == other.get_field_names()
            )

        def __hash__(self) -> int:
            return 31 * hash((self._type, self._name)) + hash(
                tuple(tuple(field_name) for field_name in self._field_names)
            )

    @final
    @dataclass(frozen=True)
    class DeleteIndex:
        """A `TableChange` to delete an index.

        If the index does not exist, the change must result in an `IllegalArgumentException`.
        """

        _name: str = field(metadata=config(field_name="name"))
        _if_exists: bool = field(metadata=config(field_name="if_exists"))

        def get_name(self) -> str:
            """Retrieves the name of the index to be deleted.

            Returns:
                str: The name of the index to be deleted.
            """
            return self._name

        def is_if_exists(self) -> bool:
            """Retrieves the value of the `if_exists` flag.

            Returns:
                bool: True if the index should be deleted if it exists, False otherwise.
            """
            return self._if_exists

        def __eq__(self, value: object) -> bool:
            if not isinstance(value, TableChange.DeleteIndex):
                return False
            other = cast(TableChange.DeleteIndex, value)
            return (
                self._name == other.get_name()
                and self._if_exists == other.is_if_exists()
            )

        def __hash__(self) -> int:
            return hash((self._name, self._if_exists))

    class ColumnPosition(ABC):
        """The interface for all column positions.

        Column positions are used to specify the position of a column when adding
        a new column to a table.
        """

        @staticmethod
        def first() -> "First":
            """The first position of `ColumnPosition` instance.

            Returns:
                First:
                    The first position of `ColumnPosition` instance.
            """
            return First()

        @staticmethod
        def after(column: str) -> "After":
            """Returns the position after the given column.

            Args:
                column:
                    The name of the reference column to place the new column after.

            Returns:
                After:
                    The position after the given column.
            """
            return After(column)

        @staticmethod
        def default_pos() -> "Default":
            """Returns the default position of `ColumnPosition` instance.

            Returns:
                Default:
                    The default position of `ColumnPosition` instance.
            """
            return Default()

    class ColumnChange(ABC):
        """The interface for all column changes.

        Column changes are used to modify the schema of a table.
        """

        @abstractmethod
        def field_name(self) -> list[str]:
            """Retrieves the field name of the column to be modified.

            Returns:
                list[str]:
                    A list of strings representing the field name.
            """


@final
@dataclass(frozen=True)
class First(TableChange.ColumnPosition):
    """Column position FIRST.

    It means the specified column should be the first column. Note that, the specified column
    may be a nested field, and then FIRST means this field should be the first one within the
    struct.
    """

    def __str__(self):
        return "FIRST"


@final
@dataclass(frozen=True)
class After(TableChange.ColumnPosition):
    """Column position AFTER

    It means the specified column should be put after the given `column`.
    Note that, the specified column may be a nested field, and then the given `column`
    refers to a field in the same struct.
    """

    _column: str = field(metadata=config(field_name="column"))

    def get_column(self) -> str:
        """Retrieves the name of the reference column after which the specified column will be placed.

        Returns:
            str: The name of the reference column.
        """
        return self._column

    def __str__(self):
        return f"AFTER {self._column}"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, After):
            return False
        other = cast(After, value)
        return self._column == other.get_column()

    def __hash__(self) -> int:
        return hash(self._column)


@final
@dataclass(frozen=True)
class Default(TableChange.ColumnPosition):
    """Column position DEFAULT.

    It means the position of the column was ignored by the user, and should be determined
    by the catalog implementation.
    """

    def __str__(self):
        return "DEFAULT"


@final
@dataclass(frozen=True)
class AddColumn(TableChange.ColumnChange):
    """A TableChange to add a field.

    The implementation may need to back-fill all the existing data to add this new column,
    or remember the column default value specified here and let the reader fill the column value
    when reading existing data that do not have this new column.

    If the field already exists, the change must result in an `IllegalArgumentException`.
    If the new field is nested and its parent does not exist or is not a struct, the change must
    result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _data_type: Type = field(metadata=config(field_name="data_type"))
    _comment: Optional[str] = field(default=None, metadata=config(field_name="comment"))
    _position: Optional[TableChange.ColumnPosition] = field(
        default=None, metadata=config(field_name="position")
    )
    _nullable: bool = field(default=True, metadata=config(field_name="nullable"))
    _auto_increment: bool = field(
        default=False, metadata=config(field_name="auto_increment")
    )
    _default_value: Optional[Expression] = field(
        default=None,
        metadata=config(field_name="default_value"),
    )

    def get_field_name(self) -> list[str]:
        """Retrieves the field name of the new column.

        Returns:
            list[str]: The field name of the new column.
        """
        return self._field_name

    def get_data_type(self) -> Type:
        """Retrieves the data type of the new column.

        Returns:
            Type: The data type of the new column.
        """
        return self._data_type

    def get_comment(self) -> Optional[str]:
        """Retrieves the comment for the new column.

        Returns:
            str: comment for the new column.
        """
        return self._comment

    def get_position(self) -> Optional[TableChange.ColumnPosition]:
        """Retrieves the position where the new column should be added.

        Returns:
            TableChange.ColumnPosition:
                The position of the column.
        """
        return self._position

    def is_nullable(self) -> bool:
        """Checks if the new column is nullable.

        Returns:
            bool: `True` if the column is nullable; `False` otherwise.
        """
        return self._nullable

    def is_auto_increment(self) -> bool:
        """Checks if the new column is auto-increment.

        Returns:
            bool: `True` if the column is auto-increment; `False` otherwise.
        """
        return self._auto_increment

    def get_default_value(self) -> Expression:
        """Retrieves the default value of the new column.

        Returns:
            Expression: The default value of the column.
        """
        return (
            self._default_value if self._default_value else Column.DEFAULT_VALUE_NOT_SET
        )

    def field_name(self) -> list[str]:
        return self._field_name

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, AddColumn):
            return False
        other = cast(AddColumn, value)
        return (
            self._nullable == other.is_nullable()
            and self._auto_increment == other.is_auto_increment()
            and self._field_name == other.get_field_name()
            and self._comment == other.get_comment()
            and self._data_type == other.get_data_type()
            and self._position == other.get_position()
            and self.get_default_value() == other.get_default_value()
        )

    def __hash__(self) -> int:
        return 31 * hash(
            (
                self._data_type,
                self._comment,
                self._position,
                self._nullable,
                self._auto_increment,
                (
                    tuple(self._default_value)
                    if self._default_value == Column.DEFAULT_VALUE_NOT_SET
                    else self._default_value
                ),
            )
        ) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class RenameColumn(TableChange.ColumnChange):
    """A `TableChange` to rename a field.

    The name is used to find the field to rename. The new name will replace the **leaf field name**.
    For example, `rename_column("a.b.c", "x")` should produce column **a.b.x**.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _new_name: str = field(metadata=config(field_name="new_name"))

    def get_field_name(self) -> list[str]:
        """Retrieves the hierarchical field name of the column to be renamed.

        Returns:
            list[str]: The field name of the column to be renamed.
        """
        return self._field_name

    def get_new_name(self) -> str:
        """Retrieves the new name for the column.

        Returns:
            str: The new name of the column.
        """
        return self._new_name

    def field_name(self) -> list[str]:
        return self._field_name

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, RenameColumn):
            return False
        other = cast(RenameColumn, value)
        return (
            self._field_name == other.get_field_name()
            and self._new_name == other.get_new_name()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._new_name) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class UpdateColumnDefaultValue(TableChange.ColumnChange):
    """A `TableChange` to update the default value of a field.

    The field names are used to find the field to update.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _new_default_value: Optional[Expression] = field(
        metadata=config(field_name="new_default_value")
    )

    def field_name(self) -> list[str]:
        return self._field_name

    def get_new_default_value(self) -> Expression:
        return (
            self._new_default_value
            if self._new_default_value
            else Column.DEFAULT_VALUE_NOT_SET
        )

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, UpdateColumnDefaultValue):
            return False
        other = cast(UpdateColumnDefaultValue, value)
        return (
            self._field_name == other.field_name()
            and self.get_new_default_value() == other.get_new_default_value()
        )

    def __hash__(self) -> int:
        return 31 * hash(
            tuple(self.get_new_default_value())
            if self._new_default_value == Column.DEFAULT_VALUE_NOT_SET
            else self._new_default_value
        ) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class UpdateColumnType(TableChange.ColumnChange):
    """A `TableChange` to update the type of a field.

    The field names are used to find the field to update.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _new_data_type: Type = field(metadata=config(field_name="new_data_type"))

    def field_name(self) -> list[str]:
        return self._field_name

    def get_field_name(self) -> list[str]:
        return self._field_name

    def get_new_data_type(self) -> Type:
        return self._new_data_type

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, UpdateColumnType):
            return False
        other = cast(UpdateColumnType, value)
        return (
            self._field_name == other.get_field_name()
            and self._new_data_type == other.get_new_data_type()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._new_data_type) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class UpdateColumnComment(TableChange.ColumnChange):
    """A `TableChange` to update the comment of a field.

    The field names are used to find the field to update.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _new_comment: str = field(metadata=config(field_name="new_comment"))

    def field_name(self) -> list[str]:
        return self._field_name

    def get_field_name(self) -> list[str]:
        return self._field_name

    def get_new_comment(self) -> str:
        return self._new_comment

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, UpdateColumnComment):
            return False
        other = cast(UpdateColumnComment, value)
        return (
            self._field_name == other.get_field_name()
            and self._new_comment == other.get_new_comment()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._new_comment) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class UpdateColumnPosition(TableChange.ColumnChange):
    """A TableChange to update the position of a field.

    The field names are used to find the field to update.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _position: TableChange.ColumnPosition = field(
        metadata=config(field_name="position")
    )

    def field_name(self) -> list[str]:
        return self._field_name

    def get_field_name(self) -> list[str]:
        """Retrieves the field name of the column whose position is being updated.

        Returns:
            list[str]: A list of strings representing the field name.
        """
        return self._field_name

    def get_position(self) -> TableChange.ColumnPosition:
        """Retrieves the new position for the column.

        Returns:
            TableChange.ColumnPosition: The new position of the column.
        """
        return self._position

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, UpdateColumnPosition):
            return False
        other = cast(UpdateColumnPosition, value)
        return (
            self._field_name == other.get_field_name()
            and self._position == other.get_position()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._position) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class DeleteColumn(TableChange.ColumnChange):
    """A TableChange to delete a field.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _if_exists: bool = field(metadata=config(field_name="if_exists"))

    def field_name(self) -> list[str]:
        return self._field_name

    def get_field_name(self) -> list[str]:
        """Retrieves the field name of the column to be deleted.

        Returns:
            list[str]: A list of strings representing the field name.
        """
        return self._field_name

    def get_if_exists(self) -> bool:
        """Checks if the field should be deleted only if it exists.

        Returns:
            bool: `True` if the field should be deleted only if it exists; `False` otherwise.
        """
        return self._if_exists

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, DeleteColumn):
            return False
        other = cast(DeleteColumn, value)
        return (
            self._field_name == other.get_field_name()
            and self._if_exists == other.get_if_exists()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._if_exists) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class UpdateColumnNullability(TableChange.ColumnChange):
    """A TableChange to update the nullability of a field.

    The field names are used to find the field to update.

    If the field does not exist, the change must result in an `IllegalArgumentException`.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _nullable: bool = field(metadata=config(field_name="nullable"))

    def field_name(self) -> list[str]:
        return self._field_name

    def get_field_name(self) -> list[str]:
        """Retrieves the field name of the column whose nullability is being updated.

        Returns:
            list[str]: A list of strings representing the field name.
        """
        return self._field_name

    def get_nullable(self) -> bool:
        """Checks if the column is nullable.

        Returns:
            bool: `True` if the column is nullable; `False` otherwise.
        """
        return self._nullable

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, UpdateColumnNullability):
            return False
        other = cast(UpdateColumnNullability, value)
        return (
            self._field_name == other.get_field_name()
            and self._nullable == other.get_nullable()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._nullable) + hash(tuple(self._field_name))


@final
@dataclass(frozen=True)
class UpdateColumnAutoIncrement(TableChange.ColumnChange):
    """A TableChange to update the `autoIncrement` of a field.

    True is to add autoIncrement, false is to delete autoIncrement.
    """

    _field_name: list[str] = field(metadata=config(field_name="field_name"))
    _auto_increment: bool = field(metadata=config(field_name="auto_increment"))

    def field_name(self) -> list[str]:
        return self._field_name

    def is_auto_increment(self) -> bool:
        """Checks if the column is `autoIncrement`.

        Returns:
            bool: `True` if the column is `autoIncrement`; `False` otherwise.
        """
        return self._auto_increment

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, UpdateColumnAutoIncrement):
            return False
        other = cast(UpdateColumnAutoIncrement, value)
        return (
            self._field_name == other.field_name()
            and self._auto_increment == other.is_auto_increment()
        )

    def __hash__(self) -> int:
        return 31 * hash(self._auto_increment) + hash(tuple(self._field_name))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from gravitino.api.rel.expressions.expression import Expression


class UnparsedExpression(Expression):
    """
    Represents an expression that is not parsed yet.
    The parsed expression is represented by FunctionExpression, literal.py, or NamedReference.
    """

    def unparsed_expression(self) -> str:
        """
        Returns the unparsed expression as a string.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def children(self) -> list[Expression]:
        """
        Unparsed expressions do not have children.
        """
        return Expression.EMPTY_EXPRESSION

    @staticmethod
    def of(unparsed_expression: str) -> UnparsedExpressionImpl:
        """
        Creates a new UnparsedExpression with the given unparsed expression.


        :param unparsed_expression: The unparsed expression as a string.
        :return: The created UnparsedExpression.
        """
        return UnparsedExpressionImpl(unparsed_expression)


class UnparsedExpressionImpl(UnparsedExpression):
    """
    An implementation of the UnparsedExpression interface.
    """

    def __init__(self, unparsed_expression: str):
        super().__init__()
        self._unparsed_expression = unparsed_expression

    def unparsed_expression(self) -> str:
        return self._unparsed_expression

    def __eq__(self, other: object) -> bool:
        if isinstance(other, UnparsedExpressionImpl):
            return self._unparsed_expression == other._unparsed_expression
        return False

    def __hash__(self) -> int:
        return hash(self._unparsed_expression)

    def __str__(self) -> str:
        """
        Returns the string representation of the unparsed expression.
        """
        return f"UnparsedExpressionImpl{{unparsedExpression='{self._unparsed_expression}'}}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from dataclasses_json import config, dataclass_json

from gravitino.rest.rest_message import RESTRequest
from gravitino.utils.precondition import Precondition


@dataclass_json
@dataclass
class TagCreateRequest(RESTRequest):
    """Represents a request to create a tag."""

    _name: str = field(metadata=config(field_name="name"))
    _comment: Optional[str] = field(default=None, metadata=config(field_name="comment"))
    _properties: Optional[dict[str, str]] = field(
        default_factory=dict, metadata=config(field_name="properties")
    )

    def validate(self) -> None:
        """
        Validate the request.
        """

        Precondition.check_string_not_empty(
            self._name, "name is required and cannot be empty"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Set

from dataclasses_json import config

from gravitino.api.model_version_change import ModelVersionChange
from gravitino.rest.rest_message import RESTRequest


@dataclass
class ModelVersionUpdateRequestBase(RESTRequest):
    """Base class for all model version update requests."""

    _type: str = field(metadata=config(field_name="@type"))

    def __init__(self, action_type: str):
        self._type = action_type

    @abstractmethod
    def model_version_change(self) -> ModelVersionChange:
        """Convert to model version change operation"""
        pass


class ModelVersionUpdateRequest:
    """Namespace for all model version update request types."""

    @dataclass
    class UpdateModelVersionComment(ModelVersionUpdateRequestBase):
        """Request to update model version comment"""

        _new_comment: Optional[str] = field(metadata=config(field_name="newComment"))
        """Represents a request to update the comment on a Metalake."""

        def __init__(self, new_comment: str):
            super().__init__("updateComment")
            self._new_comment = new_comment

        def validate(self):
            """Validates the fields of the request. Always pass."""
            pass

        def model_version_change(self):
            return ModelVersionChange.update_comment(self._new_comment)

    @dataclass
    class SetModelVersionPropertyRequest(ModelVersionUpdateRequestBase):
        """Request to update model version properties"""

        _property: Optional[str] = field(metadata=config(field_name="property"))
        _value: Optional[str] = field(metadata=config(field_name="value"))

        def __init__(self, pro: str, value: str):
            super().__init__("setProperty")
            self._property = pro
            self._value = value

        def validate(self):
            if not self._property:
                raise ValueError('"property" field is required')
            if not self._value:
                raise ValueError('"value" field is required')

        def model_version_change(self) -> ModelVersionChange:
            return ModelVersionChange.set_property(self._property, self._value)

    @dataclass
    class RemoveModelVersionPropertyRequest(ModelVersionUpdateRequestBase):
        """Request to remove model version properties"""

        _property: Optional[str] = field(metadata=config(field_name="property"))

        def __init__(self, pro: str):
            super().__init__("removeProperty")
            self._property = pro

        def key(self):
            """
            Returns the key of the property to remove.
            Returns:
                str: The key of the property to remove.
            """
            return self._property

        def validate(self):
            if not self._property:
                raise ValueError('"property" field is required')

        def model_version_change(self):
            return ModelVersionChange.remove_property(self._property)

    @dataclass
    class UpdateModelVersionUriRequest(ModelVersionUpdateRequestBase):
        """Request to update model version uri"""

        _new_uri: Optional[str] = field(metadata=config(field_name="newUri"))
        _uri_name: Optional[str] = field(metadata=config(field_name="uriName"))

        def __init__(self, new_uri: str, uri_name: str):
            super().__init__("updateUri")
            self._new_uri = new_uri
            self._uri_name = uri_name

        def new_uri(self):
            """Retrieves the new uri of the model version.
            Returns:
                The new uri of the model version.
            """
            return self._new_uri

        def uri_name(self):
            """Retrieves the uri name of the model version.
            Returns:
                The uri name of the model version.
            """
            return self._uri_name

        def validate(self):
            """Validates the fields of the request."""
            if not self._new_uri:
                raise ValueError('"newUri" field is required')

        def model_version_change(self):
            """
            Returns a ModelVersionChange object representing the update uri operation.
            Returns:
                ModelVersionChange: The ModelVersionChange object representing the update uri operation.
            """
            return ModelVersionChange.update_uri(self._new_uri, self._uri_name)

    @dataclass
    class AddModelVersionUriRequest(ModelVersionUpdateRequestBase):
        """Request to add model version uri"""

        _uri_name: Optional[str] = field(metadata=config(field_name="uriName"))
        _uri: Optional[str] = field(metadata=config(field_name="uri"))

        def __init__(self, uri_name: str, uri: str):
            super().__init__("addUri")
            self._uri_name = uri_name
            self._uri = uri

        def uri_name(self):
            """Retrieves the uri name of the model version.
            Returns:
                The uri name of the model version.
            """
            return self._uri_name

        def uri(self):
            """Retrieves the uri of the model version.
            Returns:
                The uri of the model version.
            """
            return self._uri

        def validate(self):
            """Validates the fields of the request."""
            if not self._uri_name:
                raise ValueError('"uriName" field is required')
            if not self._uri:
                raise ValueError('"uri" field is required')

        def model_version_change(self):
            """
            Returns a ModelVersionChange object representing the add uri operation.
            Returns:
                ModelVersionChange: The ModelVersionChange object representing the add uri operation.
            """
            return ModelVersionChange.add_uri(self._uri_name, self._uri)

    @dataclass
    class RemoveModelVersionUriRequest(ModelVersionUpdateRequestBase):
        """Request to remove model version uri"""

        _uri_name: Optional[str] = field(metadata=config(field_name="uriName"))

        def __init__(self, uri_name: str):
            super().__init__("removeUri")
            self._uri_name = uri_name

        def uri_name(self):
            """Retrieves the uri name of the model version.
            Returns:
                The uri name of the model version.
            """
            return self._uri_name

        def validate(self):
            """Validates the fields of the request."""
            if not self._uri_name:
                raise ValueError('"uriName" field is required')

        def model_version_change(self):
            """
            Returns a ModelVersionChange object representing the remove uri operation.
            Returns:
                ModelVersionChange: The ModelVersionChange object representing the remove uri operation.
            """
            return ModelVersionChange.remove_uri(self._uri_name)

    @dataclass
    class ModelVersionAliasesRequest(ModelVersionUpdateRequestBase):
        """Request to update model version aliases"""

        _aliases_to_add: Optional[Set[str]] = field(
            metadata=config(field_name="aliasesToAdd")
        )
        _aliases_to_remove: Optional[Set[str]] = field(
            metadata=config(field_name="aliasesToRemove")
        )

        def __init__(self, aliases_to_add: Set[str], aliases_to_remove: Set[str]):
            super().__init__("updateAliases")
            self._aliases_to_add = aliases_to_add
            self._aliases_to_remove = aliases_to_remove

        def aliases_to_add(self):
            """Retrieves the new aliases of the model version.
            Returns:
                The new aliases of the model version.
            """
            return self._aliases_to_add

        def aliases_to_remove(self):
            """Retrieves the new aliases of the model version.
            Returns:
                The new aliases of the model version.
            """
            return self._aliases_to_remove

        def validate(self):
            """Validates the fields of the request."""
            if self._aliases_to_add is None and self._aliases_to_remove is None:
                raise ValueError(
                    "At least one of aliasesToAdd or aliasesToRemove must be non-null"
                )

        def model_version_change(self):
            """
            Returns a ModelVersionChange object representing the update aliases operation.
            Returns:
                ModelVersionChange: The ModelVersionChange object representing the update aliases operation.
            """
            return ModelVersionChange.UpdateAliases(
                self._aliases_to_add,
                self._aliases_to_remove,
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field

from dataclasses_json import config


class TagChange(ABC):
    """Interface for supporting tag changes.
    This interface will be used to provide tag modification operations for each tag."""

    @staticmethod
    def rename(new_name: str) -> RenameTag:
        """
        Create a tag change instance to rename the tag.

        Args:
            new_name (str): The new name of the tag.

        Returns:
            RenameTag: A tag change instance to rename the tag.
        """
        return TagChange.RenameTag(new_name)

    @staticmethod
    def update_comment(new_comment: str) -> UpdateTagComment:
        """
        Create a tag change instance to update the tag comment.

        Args:
            new_comment (str): The new comment of the tag.

        Returns:
            UpdateTagComment: A tag change instance to update the tag comment.
        """
        return TagChange.UpdateTagComment(new_comment)

    @staticmethod
    def set_property(tag_property: str, tag_value: str) -> SetProperty:
        """
        Creates a new tag change instance to set the property and value for the tag.

        Args:
            tag_property (str): The property to set.
            tag_value (str): The value to set.

        Returns:
            SetProperty: The tag change instance to set the property and value for the tag.
        """
        return TagChange.SetProperty(tag_property, tag_value)

    @staticmethod
    def remove_property(tag_property: str) -> RemoveProperty:
        """
        Creates a new tag change instance to remove a property from the tag.

        Args:
            tag_property (str): The property to remove.

        Returns:
            RemoveProperty: The tag change instance to remove a property from the tag.
        """
        return TagChange.RemoveProperty(tag_property)

    @dataclass(frozen=True, eq=True)
    class RenameTag:
        """A tag change to rename the tag."""

        _new_name: str = field(metadata=config(field_name="newName"))

        @property
        def new_name(self) -> str:
            """The new name of the tag."""
            return self._new_name

        def __str__(self) -> str:
            return f"RENAMETAG {self._new_name}"

    @dataclass(frozen=True, eq=True)
    class UpdateTagComment:
        """A tag change to update the tag comment."""

        _new_comment: str = field(metadata=config(field_name="newComment"))

        @property
        def new_comment(self) -> str:
            """The new comment of the tag."""
            return self._new_comment

        def __str__(self) -> str:
            return f"UPDATETAGCOMMENT {self._new_comment}"

    @dataclass(frozen=True, eq=True)
    class SetProperty:
        """A tag change to set a property-value pair for the tag."""

        _property: str = field(metadata=config(field_name="property"))
        _value: str = field(metadata=config(field_name="value"))

        @property
        def name(self) -> str:
            """The property to set."""
            return self._property

        @property
        def value(self) -> str:
            """The value to set."""
            return self._value

        def __str__(self) -> str:
            return f"SETTAGPROPERTY {self._property} = {self._value}"

    @dataclass(frozen=True, eq=True)
    class RemoveProperty:
        """A tag change to remove a property-value pair for the tag."""

        _property: str = field(metadata=config(field_name="property"))

        @property
        def removed_property(self) -> str:
            """The property to remove."""
            return self._property

        def __str__(self) -> str:
            return f"REMOVETAGPROPERTY {self._property}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements. See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership. The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied. See the License for the
# specific language governing permissions and limitations
# under the License.

from copy import deepcopy
from typing import List, Any
from .job_template import JobTemplate, JobType


class ShellJobTemplate(JobTemplate):
    """
    Represents a job template for executing shell scripts.
    """

    def __init__(self, builder: "ShellJobTemplate.Builder"):
        """
        Initializes a ShellJobTemplate instance.
        """
        super().__init__(builder)
        self._scripts = builder.scripts

    @property
    def scripts(self) -> List[str]:
        """
        Returns the list of shell scripts associated with this job template.

        Returns:
            List of shell scripts.
        """
        return self._scripts

    def job_type(self) -> JobType:
        """
        Returns the type of the job, which is SHELL for this template.

        Returns:
            JobType.SHELL: Indicates that this is a shell job template.
        """
        return JobType.SHELL

    @staticmethod
    def builder() -> "ShellJobTemplate.Builder":
        """
        Creates a new builder instance for constructing a ShellJobTemplate.

        Returns:
            ShellJobTemplate.Builder: A new builder instance.
        """
        return ShellJobTemplate.Builder()

    def __eq__(self, other: Any) -> bool:
        if not super().__eq__(other):
            return False
        return self._scripts == other.scripts

    def __hash__(self) -> int:
        return hash((super().__hash__(), tuple(self._scripts)))

    def __str__(self) -> str:
        res = "\nSellJobTemplate{\n"
        if self._scripts:
            res += (
                "  scripts=[\n"
                + "".join(f"    {s},\n" for s in self._scripts)
                + "  ],\n"
            )
        else:
            res += "  scripts=[],\n"

        return res + super().__str__() + "}\n"

    class Builder(
        JobTemplate.BaseBuilder["ShellJobTemplate.Builder", "ShellJobTemplate"]
    ):
        """
        Builder class for creating ShellJobTemplate instances.
        """

        def __init__(self):
            """
            Initializes the builder with default values.
            """
            super().__init__()
            self.scripts: List[str] = []

        def with_scripts(self, scripts: List[str]) -> "ShellJobTemplate.Builder":
            """
            Sets the shell scripts for this job template.

            Args:
                scripts: A list of shell scripts to be executed by this job template.

            Returns:
                ShellJobTemplate.Builder: The builder instance for method chaining.
            """
            self.scripts = deepcopy(scripts)
            return self

        def validate(self):
            """
            Validates the builder state before building the ShellJobTemplate.
            """
            super().validate()
            self.scripts = self.scripts or []

        def build(self) -> "ShellJobTemplate":
            """
            Builds and returns a ShellJobTemplate instance.

            Returns:
                ShellJobTemplate: The constructed shell job template.
            """
            self.validate()
            return ShellJobTemplate(self)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import re
from collections.abc import Mapping
from types import MappingProxyType
from typing import Final, Pattern, Set

from gravitino.api.rel.types.types import Name, Types


class SerdesUtilsBase:
    EXPRESSION_TYPE: Final[str] = "type"
    DATA_TYPE: Final[str] = "dataType"
    LITERAL_VALUE: Final[str] = "value"
    FIELD_NAME: Final[str] = "fieldName"
    FUNCTION_NAME: Final[str] = "funcName"
    FUNCTION_ARGS: Final[str] = "funcArgs"
    UNPARSED_EXPRESSION: Final[str] = "unparsedExpression"
    TYPE: Final[str] = "type"
    STRUCT: Final[str] = "struct"
    FIELDS: Final[str] = "fields"
    STRUCT_FIELD_NAME: Final[str] = "name"
    STRUCT_FIELD_NULLABLE: Final[str] = "nullable"
    STRUCT_FIELD_COMMENT: Final[str] = "comment"
    LIST: Final[str] = "list"
    LIST_ELEMENT_TYPE: Final[str] = "elementType"
    LIST_ELEMENT_NULLABLE: Final[str] = "containsNull"
    MAP: Final[str] = "map"
    MAP_KEY_TYPE: Final[str] = "keyType"
    MAP_VALUE_TYPE: Final[str] = "valueType"
    MAP_VALUE_NULLABLE: Final[str] = "valueContainsNull"
    UNION: Final[str] = "union"
    UNION_TYPES: Final[str] = "types"
    UNPARSED: Final[str] = "unparsed"
    UNPARSED_TYPE: Final[str] = "unparsedType"
    EXTERNAL: Final[str] = "external"
    CATALOG_STRING: Final[str] = "catalogString"

    PARTITION_TYPE: Final[str] = "type"
    PARTITION_NAME: Final[str] = "name"
    PARTITION_PROPERTIES: Final[str] = "properties"
    FIELD_NAMES: Final[str] = "fieldNames"
    IDENTITY_PARTITION_VALUES: Final[str] = "values"
    LIST_PARTITION_LISTS: Final[str] = "lists"
    RANGE_PARTITION_UPPER: Final[str] = "upper"
    RANGE_PARTITION_LOWER: Final[str] = "lower"
    STRATEGY: Final[str] = "strategy"
    NUM_BUCKETS: Final[str] = "numBuckets"
    WIDTH: Final[str] = "width"
    ASSIGNMENTS_NAME: Final[str] = "assignments"
    SORT_TERM: Final[str] = "sortTerm"
    DIRECTION: Final[str] = "direction"
    NULL_ORDERING: Final[str] = "nullOrdering"
    INDEX_TYPE: Final[str] = "indexType"
    INDEX_NAME: Final[str] = "name"
    INDEX_FIELD_NAMES: Final[str] = "fieldNames"
    NUMBER: Final[str] = "number"

    NON_PRIMITIVE_TYPES: Final[Set[Name]] = {
        Name.STRUCT,
        Name.LIST,
        Name.MAP,
        Name.UNION,
        Name.UNPARSED,
        Name.EXTERNAL,
    }
    PRIMITIVE_AND_NULL_TYPES: Final[Set[Name]] = set(list(Name)) - NON_PRIMITIVE_TYPES

    DECIMAL_PATTERN: Final[Pattern[str]] = re.compile(
        r"decimal\(\s*(\d+)\s*,\s*(\d+)\s*\)"
    )
    FIXED_PATTERN: Final[Pattern[str]] = re.compile(r"fixed\(\s*(\d+)\s*\)")
    FIXEDCHAR_PATTERN: Final[Pattern[str]] = re.compile(r"char\(\s*(\d+)\s*\)")
    VARCHAR_PATTERN: Final[Pattern[str]] = re.compile(r"varchar\(\s*(\d+)\s*\)")
    TIME_PATTERN: Final[Pattern[str]] = re.compile(r"time\(\s*(\d+)\s*\)")
    TIMESTAMP_PATTERN: Final[Pattern[str]] = re.compile(r"timestamp\(\s*(\d+)\s*\)")
    TIMESTAMP_TZ_PATTERN: Final[Pattern[str]] = re.compile(
        r"timestamp_tz\(\s*(\d+)\s*\)"
    )
    TYPES: Final[Mapping] = MappingProxyType(
        {
            type_instance.simple_string(): type_instance
            for type_instance in (
                Types.NullType.get(),
                Types.BooleanType.get(),
                Types.ByteType.get(),
                Types.ByteType.unsigned(),
                Types.IntegerType.get(),
                Types.IntegerType.unsigned(),
                Types.ShortType.get(),
                Types.ShortType.unsigned(),
                Types.LongType.get(),
                Types.LongType.unsigned(),
                Types.FloatType.get(),
                Types.DoubleType.get(),
                Types.DateType.get(),
                Types.TimeType.get(),
                Types.TimestampType.with_time_zone(),
                Types.TimestampType.without_time_zone(),
                Types.IntervalYearType.get(),
                Types.IntervalDayType.get(),
                Types.StringType.get(),
                Types.UUIDType.get(),
            )
        }
    )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC
from typing import List

from gravitino.api.function.function_definition import FunctionDefinition
from gravitino.api.function.function_impl import FunctionImpl
from gravitino.api.function.function_param import FunctionParam


class FunctionChange(ABC):
    """Represents a change that can be applied to a function."""

    EMPTY_PARAMS: List[FunctionParam] = []
    """An empty list of parameters."""

    @staticmethod
    def update_comment(new_comment: str) -> "FunctionChange":
        """Create a FunctionChange to update the comment of a function.

        Args:
            new_comment: The new comment value.

        Returns:
            The change instance.
        """
        return UpdateComment(new_comment)

    @staticmethod
    def add_definition(definition: FunctionDefinition) -> "FunctionChange":
        """Create a FunctionChange to add a new definition (overload) to a function.

        Args:
            definition: The new definition to add.

        Returns:
            The change instance.
        """
        return AddDefinition(definition)

    @staticmethod
    def remove_definition(parameters: List[FunctionParam]) -> "FunctionChange":
        """Create a FunctionChange to remove an existing definition from a function.

        Args:
            parameters: The parameters that identify the definition to remove.

        Returns:
            The change instance.
        """
        return RemoveDefinition(parameters)

    @staticmethod
    def add_impl(
        parameters: List[FunctionParam], implementation: FunctionImpl
    ) -> "FunctionChange":
        """Create a FunctionChange to add an implementation to a specific definition.

        Args:
            parameters: The parameters that identify the definition to update.
            implementation: The implementation to add.

        Returns:
            The change instance.
        """
        return AddImpl(parameters, implementation)

    @staticmethod
    def update_impl(
        parameters: List[FunctionParam],
        runtime: FunctionImpl.RuntimeType,
        implementation: FunctionImpl,
    ) -> "FunctionChange":
        """Create a FunctionChange to update an implementation for a specific definition.

        Args:
            parameters: The parameters that identify the definition to update.
            runtime: The runtime that identifies the implementation to replace.
            implementation: The new implementation.

        Returns:
            The change instance.
        """
        return UpdateImpl(parameters, runtime, implementation)

    @staticmethod
    def remove_impl(
        parameters: List[FunctionParam], runtime: FunctionImpl.RuntimeType
    ) -> "FunctionChange":
        """Create a FunctionChange to remove an implementation for a specific definition.

        Args:
            parameters: The parameters that identify the definition to update.
            runtime: The runtime that identifies the implementation to remove.

        Returns:
            The change instance.
        """
        return RemoveImpl(parameters, runtime)


class UpdateComment(FunctionChange):
    """A FunctionChange to update the comment of a function."""

    def __init__(self, new_comment: str):
        if not new_comment or not new_comment.strip():
            raise ValueError("New comment cannot be null or empty")
        self._new_comment = new_comment

    def new_comment(self) -> str:
        """Returns the new comment of the function."""
        return self._new_comment

    def __eq__(self, other) -> bool:
        if not isinstance(other, UpdateComment):
            return False
        return self._new_comment == other._new_comment

    def __hash__(self) -> int:
        return hash(self._new_comment)

    def __repr__(self) -> str:
        return f"UpdateComment(newComment='{self._new_comment}')"


class AddDefinition(FunctionChange):
    """A FunctionChange to add a new definition to a function."""

    def __init__(self, definition: FunctionDefinition):
        if definition is None:
            raise ValueError("Definition cannot be null")
        self._definition = definition

    def definition(self) -> FunctionDefinition:
        """Returns the definition to add."""
        return self._definition

    def __eq__(self, other) -> bool:
        if not isinstance(other, AddDefinition):
            return False
        return self._definition == other._definition

    def __hash__(self) -> int:
        return hash(self._definition)

    def __repr__(self) -> str:
        return f"AddDefinition(definition={self._definition})"


class RemoveDefinition(FunctionChange):
    """A FunctionChange to remove an existing definition from a function."""

    def __init__(self, parameters: List[FunctionParam]):
        if parameters is None:
            raise ValueError("Parameters cannot be null")
        self._parameters = list(parameters)

    def parameters(self) -> List[FunctionParam]:
        """Returns the parameters that identify the definition to remove."""
        return (
            list(self._parameters) if self._parameters else FunctionChange.EMPTY_PARAMS
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, RemoveDefinition):
            return False
        return self._parameters == other._parameters

    def __hash__(self) -> int:
        return hash(tuple(self._parameters))

    def __repr__(self) -> str:
        return f"RemoveDefinition(parameters={self._parameters})"


class AddImpl(FunctionChange):
    """A FunctionChange to add an implementation to a definition."""

    def __init__(self, parameters: List[FunctionParam], implementation: FunctionImpl):
        if parameters is None:
            raise ValueError("Parameters cannot be null")
        self._parameters = list(parameters)
        if implementation is None:
            raise ValueError("Implementation cannot be null")
        self._implementation = implementation

    def parameters(self) -> List[FunctionParam]:
        """Returns the parameters that identify the definition to update."""
        return (
            list(self._parameters) if self._parameters else FunctionChange.EMPTY_PARAMS
        )

    def implementation(self) -> FunctionImpl:
        """Returns the implementation to add."""
        return self._implementation

    def __eq__(self, other) -> bool:
        if not isinstance(other, AddImpl):
            return False
        return (
            self._parameters == other._parameters
            and self._implementation == other._implementation
        )

    def __hash__(self) -> int:
        return hash((tuple(self._parameters), self._implementation))

    def __repr__(self) -> str:
        return (
            f"AddImpl(parameters={self._parameters}, "
            f"implementation={self._implementation})"
        )


class UpdateImpl(FunctionChange):
    """A FunctionChange to replace an implementation for a specific definition."""

    def __init__(
        self,
        parameters: List[FunctionParam],
        runtime: FunctionImpl.RuntimeType,
        implementation: FunctionImpl,
    ):
        if parameters is None:
            raise ValueError("Parameters cannot be null")
        self._parameters = list(parameters)
        if runtime is None:
            raise ValueError("Runtime cannot be null")
        self._runtime = runtime
        if implementation is None:
            raise ValueError("Implementation cannot be null")
        self._implementation = implementation
        if runtime != implementation.runtime():
            raise ValueError(
                "Runtime of implementation must match the runtime being updated"
            )

    def parameters(self) -> List[FunctionParam]:
        """Returns the parameters that identify the definition to update."""
        return (
            list(self._parameters) if self._parameters else FunctionChange.EMPTY_PARAMS
        )

    def runtime(self) -> FunctionImpl.RuntimeType:
        """Returns the runtime that identifies the implementation to replace."""
        return self._runtime

    def implementation(self) -> FunctionImpl:
        """Returns the new implementation."""
        return self._implementation

    def __eq__(self, other) -> bool:
        if not isinstance(other, UpdateImpl):
            return False
        return (
            self._parameters == other._parameters
            and self._runtime == other._runtime
            and self._implementation == other._implementation
        )

    def __hash__(self) -> int:
        return hash((tuple(self._parameters), self._runtime, self._implementation))

    def __repr__(self) -> str:
        return (
            f"UpdateImpl(parameters={self._parameters}, runtime={self._runtime}, "
            f"implementation={self._implementation})"
        )


class RemoveImpl(FunctionChange):
    """A FunctionChange to remove an implementation for a specific runtime."""

    def __init__(
        self, parameters: List[FunctionParam], runtime: FunctionImpl.RuntimeType
    ):
        if parameters is None:
            raise ValueError("Parameters cannot be null")
        self._parameters = list(parameters)
        if runtime is None:
            raise ValueError("Runtime cannot be null")
        self._runtime = runtime

    def parameters(self) -> List[FunctionParam]:
        """Returns the parameters that identify the definition to update."""
        return (
            list(self._parameters) if self._parameters else FunctionChange.EMPTY_PARAMS
        )

    def runtime(self) -> FunctionImpl.RuntimeType:
        """Returns the runtime that identifies the implementation to remove."""
        return self._runtime

    def __eq__(self, other) -> bool:
        if not isinstance(other, RemoveImpl):
            return False
        return self._parameters == other._parameters and self._runtime == other._runtime

    def __hash__(self) -> int:
        return hash((tuple(self._parameters), self._runtime))

    def __repr__(self) -> str:
        return f"RemoveImpl(parameters={self._parameters}, runtime={self._runtime})"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import threading
from typing import Dict

caller_context_holder = threading.local()


class CallerContext:
    """A class defining the caller context for auditing coarse-grained operations."""

    _context: Dict[str, str]

    def __init__(self, context: Dict[str, str]):
        """Initialize the CallerContext.

        Args:
            context: The context dict.
        """
        self._context = context

    def context(self) -> Dict[str, str]:
        """Returns the context dict in the caller context.

        Returns:
             The context dict.
        """
        return self._context


class CallerContextHolder:
    """A thread local holder for the CallerContext."""

    @staticmethod
    def set(context: CallerContext):
        """Set the CallerContext in the thread local.

        Args:
             context: The CallerContext to set.
        """
        caller_context_holder.caller_context = context

    @staticmethod
    def get():
        """Get the CallerContext from the thread local.

        Returns:
             The CallerContext.
        """
        if not hasattr(caller_context_holder, "caller_context"):
            return None
        return caller_context_holder.caller_context

    @staticmethod
    def remove():
        """Remove the CallerContext from the thread local."""
        if hasattr(caller_context_holder, "caller_context"):
            del caller_context_holder.caller_context



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any

from gravitino.api.rel.indexes.index import Index
from gravitino.api.rel.types.json_serdes import JsonSerializable
from gravitino.dto.rel.indexes.index_dto import IndexDTO
from gravitino.utils.precondition import Precondition
from gravitino.utils.serdes import SerdesUtilsBase


class IndexSerdes(SerdesUtilsBase, JsonSerializable[Index]):
    @classmethod
    def serialize(cls, data_type: Index) -> dict[str, Any]:
        result: dict[str, Any] = {cls.INDEX_TYPE: data_type.type().name.upper()}
        if data_type.name() is not None:
            result[cls.INDEX_NAME] = data_type.name()
        result[cls.INDEX_FIELD_NAMES] = data_type.field_names()

        return result

    @classmethod
    def deserialize(cls, data: dict[str, Any]) -> Index:
        Precondition.check_argument(
            isinstance(data, dict) and len(data) > 0,
            f"Index must be a valid JSON object, but found: {data}",
        )
        Precondition.check_argument(
            data.get(cls.INDEX_TYPE) is not None,
            f"Cannot parse index from missing type: {data}",
        )
        Precondition.check_argument(
            data.get(cls.INDEX_FIELD_NAMES) is not None,
            f"Cannot parse index from missing field names: {data}",
        )
        index_type = Index.IndexType(data[cls.INDEX_TYPE].upper())

        return IndexDTO(
            index_type, data.get(cls.INDEX_NAME), data[cls.INDEX_FIELD_NAMES]
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Dict

from gravitino.api.rel.partitions.range_partition import RangePartition
from gravitino.dto.rel.expressions.literal_dto import LiteralDTO
from gravitino.dto.rel.partitions.partition_dto import PartitionDTO


class RangePartitionDTO(PartitionDTO, RangePartition):
    """Data transfer object representing a range partition."""

    def __init__(
        self,
        name: str,
        properties: Dict[str, str],
        upper: LiteralDTO,
        lower: LiteralDTO,
    ):
        self._name = name
        self._properties = properties
        self._upper = upper
        self._lower = lower

    def name(self) -> str:
        return self._name

    def properties(self) -> Dict[str, str]:
        return self._properties

    def upper(self) -> LiteralDTO:
        return self._upper

    def lower(self) -> LiteralDTO:
        return self._lower

    def type(self) -> PartitionDTO.Type:
        return self.Type.RANGE

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, RangePartitionDTO):
            return False

        return (
            self is value
            or self._name == value.name()
            and self._properties == value.properties()
            and self._upper == value.upper()
            and self._lower == value.lower()
        )

    def __hash__(self) -> int:
        return hash(
            (self._name, tuple(self._properties.items()), self._upper, self._lower)
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Dict, List, Optional

from gravitino.api.catalog import Catalog
from gravitino.api.catalog_change import CatalogChange
from gravitino.api.job.job_handle import JobHandle
from gravitino.api.job.job_template import JobTemplate
from gravitino.api.job.job_template_change import JobTemplateChange
from gravitino.api.job.supports_jobs import SupportsJobs
from gravitino.api.tag.tag_operations import TagOperations
from gravitino.auth.auth_data_provider import AuthDataProvider
from gravitino.client.gravitino_client_base import GravitinoClientBase
from gravitino.client.gravitino_metalake import GravitinoMetalake

from ..api.tag.tag import Tag


class GravitinoClient(GravitinoClientBase, SupportsJobs, TagOperations):
    """Gravitino Client for a user to interact with the Gravitino API, allowing the client to list,
    load, create, and alter Catalog.

    It uses an underlying {@link RESTClient} to send HTTP requests and receive responses from the API.
    """

    _metalake: GravitinoMetalake

    def __init__(
        self,
        uri: str,
        metalake_name: str,
        check_version: bool = True,
        auth_data_provider: AuthDataProvider = None,
        request_headers: Optional[dict] = None,
        client_config: Optional[dict] = None,
    ):
        """Constructs a new GravitinoClient with the given URI, authenticator and AuthDataProvider.

        Args:
            uri: The base URI for the Gravitino API.
            metalake_name: The specified metalake name.
            auth_data_provider: The provider of the data which is used for authentication.
            request_headers: The headers to be included in the HTTP requests.
            client_config: The config properties for the HTTP Client

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.
        """
        super().__init__(
            uri, check_version, auth_data_provider, request_headers, client_config
        )
        self.check_metalake_name(metalake_name)
        self._metalake = super().load_metalake(metalake_name)

    def get_metalake(self) -> GravitinoMetalake:
        """Get the current metalake object

        Raises:
            NoSuchMetalakeException if the metalake with specified name does not exist.

        Returns:
            the GravitinoMetalake object
        """
        return self._metalake

    def list_catalogs(self) -> List[str]:
        return self.get_metalake().list_catalogs()

    def list_catalogs_info(self) -> List[Catalog]:
        return self.get_metalake().list_catalogs_info()

    def load_catalog(self, name: str) -> Catalog:
        return self.get_metalake().load_catalog(name)

    def create_catalog(
        self,
        name: str,
        catalog_type: Catalog.Type,
        provider: str,
        comment: str,
        properties: Dict[str, str],
    ) -> Catalog:
        return self.get_metalake().create_catalog(
            name, catalog_type, provider, comment, properties
        )

    def alter_catalog(self, name: str, *changes: CatalogChange):
        return self.get_metalake().alter_catalog(name, *changes)

    def drop_catalog(self, name: str, force: bool = False) -> bool:
        return self.get_metalake().drop_catalog(name, force)

    def enable_catalog(self, name: str):
        return self.get_metalake().enable_catalog(name)

    def disable_catalog(self, name: str):
        return self.get_metalake().disable_catalog(name)

    def list_job_templates(self) -> List[JobTemplate]:
        """Lists all job templates in the current metalake.

        Returns:
            A list of JobTemplate objects representing the job templates in the metalake.
        """
        return self.get_metalake().list_job_templates()

    def register_job_template(self, job_template) -> None:
        """Register a job template with the specified job template to Gravitino. The registered
        job template will be maintained in Gravitino, allowing it to be executed later.

        Args:
            job_template: The job template to register.

        Raises:
            JobTemplateAlreadyExists: If a job template with the same name already exists.
        """
        self.get_metalake().register_job_template(job_template)

    def get_job_template(self, job_template_name: str) -> JobTemplate:
        """Retrieves a job template by its name.

        Args:
            job_template_name: The name of the job template to retrieve.

        Returns:
            The JobTemplate object corresponding to the specified name.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
        """
        return self.get_metalake().get_job_template(job_template_name)

    def delete_job_template(self, job_template_name: str) -> bool:
        """
        Deletes a job template by its name. This will remove the job template from Gravitino, and it
        will no longer be available for execution. Only when all the jobs associated with this job
        template are completed, failed or cancelled, the job template can be deleted successfully,
        otherwise it will throw InUseException. Returns false if the job template to be deleted does
        not exist.

        The deletion of a job template will also delete all the jobs associated with this template.

        Args:
            job_template_name: The name of the job template to delete.

        Returns:
            bool: True if the job template was deleted successfully, False if the job template
            does not exist.

        Raises:
            InUseException: If there are still queued or started jobs associated with this job template.
        """
        return self.get_metalake().delete_job_template(job_template_name)

    def alter_job_template(
        self, job_template_name: str, *changes: JobTemplateChange
    ) -> JobTemplate:
        """Alters a job template with the specified changes.

        Args:
            job_template_name: The name of the job template to alter.
            changes: The changes to apply to the job template.

        Returns:
            The altered JobTemplate object.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
            IllegalArgumentException: If any of the changes cannot be applied.
        """
        return self.get_metalake().alter_job_template(job_template_name, *changes)

    def list_jobs(self, job_template_name: str = None) -> List[JobHandle]:
        """Lists all the jobs in the current metalake.

        Args:
            job_template_name: Optional; if provided, filters the jobs by the specified job template name.

        Returns:
            A list of JobHandle objects representing the jobs in the metalake.
        """
        return self.get_metalake().list_jobs(job_template_name)

    def get_job(self, job_id: str) -> JobHandle:
        """Retrieves a job by its ID.

        Args:
            job_id: The ID of the job to retrieve.

        Returns:
            The JobHandle object corresponding to the specified job ID.

        Raises:
            NoSuchJobException: If no job with the specified ID exists.
        """
        return self.get_metalake().get_job(job_id)

    def run_job(self, job_template_name: str, job_conf: Dict[str, str]) -> JobHandle:
        """Runs a job using the specified job template and configuration.

        Args:
            job_template_name: The name of the job template to use for running the job.
            job_conf: A dictionary containing the configuration for the job.

        Returns:
            A JobHandle object representing the started job.

        Raises:
            NoSuchJobTemplateException: If no job template with the specified name exists.
        """
        return self.get_metalake().run_job(job_template_name, job_conf)

    def cancel_job(self, job_id: str) -> JobHandle:
        """Cancels a job by its ID.

        Args:
            job_id: The ID of the job to cancel.

        Returns:
            The JobHandle object representing the cancelled job.

        Raises:
            NoSuchJobException: If no job with the specified ID exists.
        """
        return self.get_metalake().cancel_job(job_id)

    # Tag operations
    def list_tags(self) -> list[str]:
        """List all the tag names under a metalake.

        Returns:
            list[str]: The list of tag names.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
        """
        return self.get_metalake().list_tags()

    def list_tags_info(self) -> list[Tag]:
        """
        List tags information under a metalake.

        Returns:
            list[Tag]: The list of tag information.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
        """
        return self.get_metalake().list_tags_info()

    def get_tag(self, tag_name) -> Tag:
        """
        Get a tag by its name under a metalake.

        Args:
            tag_name (str): The name of the tag.

        Returns:
            Tag: The tag information.

        Raises:
            NoSuchTagException: If the tag does not exist.
        """
        return self.get_metalake().get_tag(tag_name)

    def create_tag(self, tag_name, comment, properties) -> Tag:
        """
        Create a new tag under a metalake.

        Raises:
            NoSuchMetalakeException: If the metalake does not exist.
            TagAlreadyExistsException: If the tag already exists.

        Args:
            tag_name (str): The name of the tag.
            comment (str): The comment of the tag.
            properties (dict[str, str]): The properties of the tag.

        Returns:
            Tag: The tag information.
        """
        return self.get_metalake().create_tag(tag_name, comment, properties)

    def alter_tag(self, tag_name, *changes) -> Tag:
        """
        Alter a tag under a metalake.

        Args:
            tag_name (str): The name of the tag.
            changes (TagChange): The changes to apply to the tag.

        Returns:
            Tag: The altered tag.

        Raises:
            NoSuchTagException: If the tag does not exist.
            NoSuchMetalakeException: If the metalake does not exist.
        """
        return self.get_metalake().alter_tag(tag_name, *changes)

    def delete_tag(self, tag_name) -> bool:
        """
        Delete a tag under a metalake.

        Args:
            tag_name (str): The name of the tag.

        Returns:
            bool: True if the tag was deleted, False otherwise.

        Raises:
            NoSuchTagException: If the tag does not exist.
            NoSuchMetalakeException: If the metalake does not exist.
        """
        return self.get_metalake().delete_tag(tag_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Optional

from gravitino.api.rel.types.type import Type


class FunctionColumn:
    """Represents a return column of a table-valued function."""

    def __init__(self, name: str, data_type: Type, comment: Optional[str] = None):
        """Create a FunctionColumn instance.

        Args:
            name: The column name.
            data_type: The column type.
            comment: The optional comment of the column.

        Raises:
            ValueError: If name is null or empty, or data_type is null.
        """
        if not name or not name.strip():
            raise ValueError("Function column name cannot be null or empty")
        self._name = name

        if data_type is None:
            raise ValueError("Function column type cannot be null")
        self._data_type = data_type

        self._comment = comment

    @classmethod
    def of(
        cls, name: str, data_type: Type, comment: Optional[str] = None
    ) -> "FunctionColumn":
        """Create a FunctionColumn instance.

        Args:
            name: The column name.
            data_type: The column type.
            comment: The optional comment of the column.

        Returns:
            A FunctionColumn instance.
        """
        return cls(name, data_type, comment)

    def name(self) -> str:
        """Returns the column name."""
        return self._name

    def data_type(self) -> Type:
        """Returns the column type."""
        return self._data_type

    def comment(self) -> Optional[str]:
        """Returns the optional column comment, None if not provided."""
        return self._comment

    def __eq__(self, other) -> bool:
        if not isinstance(other, FunctionColumn):
            return False
        return (
            self._name == other._name
            and self._data_type == other._data_type
            and self._comment == other._comment
        )

    def __hash__(self) -> int:
        return hash((self._name, self._data_type, self._comment))

    def __repr__(self) -> str:
        return (
            f"FunctionColumn(name='{self._name}', dataType={self._data_type}, "
            f"comment='{self._comment}')"
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from typing import Optional

from gravitino.api.metadata_object import MetadataObject
from gravitino.api.tag.tag import Tag
from gravitino.dto.audit_dto import AuditDTO
from gravitino.dto.responses.metadata_object_list_response import (
    MetadataObjectListResponse,
)
from gravitino.dto.tag_dto import TagDTO
from gravitino.exceptions.handlers.error_handler import ErrorHandler
from gravitino.exceptions.handlers.tag_error_handler import TAG_ERROR_HANDLER
from gravitino.rest.rest_utils import encode_string
from gravitino.utils import HTTPClient
from gravitino.utils.http_client import Response


class GenericTag(Tag, Tag.AssociatedObjects):
    """Represents a generic tag."""

    API_LIST_OBJECTS_ENDPOINT = "api/metalakes/{}/tags/{}/objects"

    def __init__(
        self,
        metalake: str,
        tag_dto: TagDTO,
        client: HTTPClient,
    ) -> None:
        self._metalake = metalake
        self._tag_dto = tag_dto
        self._client = client

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GenericTag):
            return False

        return self._tag_dto == other._tag_dto

    def __hash__(self) -> int:
        return hash(self._tag_dto)

    def name(self) -> str:
        """Get the name of the tag.

        Returns:
            str: The name of the tag.
        """
        return self._tag_dto.name()

    def comment(self) -> str:
        """Get the comment of the tag.

        Returns:
            str: The comment of the tag.
        """
        return self._tag_dto.comment()

    def properties(self) -> dict[str, str]:
        """Get the properties of the tag.

        Returns:
            Dict[str, str]: The properties of the tag.
        """
        return self._tag_dto.properties()

    def inherited(self) -> Optional[bool]:
        """Check if the tag is inherited from a parent object or not.

        If the tag is inherited, it will return `True`, if it is owned by the object itself, it will return `False`.

        **Note**. The return value is optional, only when the tag is associated with an object, and called from the
        object, the return value will be present. Otherwise, it will be empty.

        Returns:
            Optional[bool]:
                True if the tag is inherited, false if it is owned by the object itself. Empty if the
                tag is not associated with any object.
        """
        return self._tag_dto.inherited()

    def audit_info(self) -> AuditDTO:
        """
        Retrieve the audit information of the entity.

        Returns:
            AuditDTO: The audit information of the entity.
        """
        return self._tag_dto.audit_info()

    def associated_objects(self) -> Tag.AssociatedObjects:
        """
        The associated objects of the tag.

        Returns:
            AssociatedObjects: The associated objects of the tag.
        """
        return self

    def objects(self) -> list[MetadataObject]:
        """
        Retrieve the list of objects that are associated with this tag.

        Returns:
            list[MetadataObject]: The list of objects that are associated with this tag.
        """
        url = self.API_LIST_OBJECTS_ENDPOINT.format(
            self._metalake,
            encode_string(self.name()),
        )

        response = self.get_response(url, TAG_ERROR_HANDLER)
        objects_resp = MetadataObjectListResponse.from_json(
            response.body, infer_missing=True
        )
        objects_resp.validate()

        return objects_resp.metadata_objects()

    def get_response(self, url: str, error_handler: ErrorHandler) -> Response:
        """
        Get the response from the server, for testing convenience.

        Args:
            url (str): The url to get the response from.
            error_handler (ErrorHandlers): The error handler to use.

        Returns:
            Response: The response from the server.
        """
        return self._client.get(
            url,
            error_handler=error_handler,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABC, abstractmethod
from typing import Optional

from gravitino.api.rel.expressions.expression import Expression
from gravitino.api.rel.types.type import Type


class FunctionParam(ABC):
    """Represents a function parameter."""

    @abstractmethod
    def name(self) -> str:
        """Returns the name of the parameter."""
        pass

    @abstractmethod
    def data_type(self) -> Type:
        """Returns the data type of the parameter."""
        pass

    def comment(self) -> Optional[str]:
        """Returns the optional comment of the parameter, None if not provided."""
        return None

    def default_value(self) -> Optional[Expression]:
        """Returns the default value of the parameter if provided, otherwise None."""
        return None


class FunctionParams:
    """Factory class for creating FunctionParam instances."""

    class SimpleFunctionParam(FunctionParam):
        """Simple implementation of FunctionParam."""

        def __init__(
            self,
            name: str,
            data_type: Type,
            comment: Optional[str] = None,
            default_value: Optional[Expression] = None,
        ):
            if not name or not name.strip():
                raise ValueError("Parameter name cannot be null or empty")
            self._name = name

            if data_type is None:
                raise ValueError("Parameter data type cannot be null")
            self._data_type = data_type

            self._comment = comment
            self._default_value = default_value

        def name(self) -> str:
            return self._name

        def data_type(self) -> Type:
            return self._data_type

        def comment(self) -> Optional[str]:
            return self._comment

        def default_value(self) -> Optional[Expression]:
            return self._default_value

        def __eq__(self, other) -> bool:
            if not isinstance(other, FunctionParam):
                return False
            return (
                self._name == other.name()
                and self._data_type == other.data_type()
                and self._comment == other.comment()
                and self._default_value == other.default_value()
            )

        def __hash__(self) -> int:
            return hash(
                (self._name, self._data_type, self._comment, self._default_value)
            )

        def __repr__(self) -> str:
            return (
                f"FunctionParam(name='{self._name}', dataType={self._data_type}, "
                f"comment='{self._comment}', defaultValue={self._default_value})"
            )

    @classmethod
    def of(
        cls,
        name: str,
        data_type: Type,
        comment: Optional[str] = None,
        default_value: Optional[Expression] = None,
    ) -> FunctionParam:
        """Create a FunctionParam instance.

        Args:
            name: The parameter name.
            data_type: The parameter type.
            comment: The optional comment.
            default_value: The optional default value.

        Returns:
            A FunctionParam instance.
        """
        return cls.SimpleFunctionParam(name, data_type, comment, default_value)
